"""This module contains class representations for XML nodes.

For the purpose of storing the validation results, we would like to reference the XML node at
which a check was performed. The ISO Schematron applies rules to:

- Elements (*)
- Attributes (@*)
- Root node (/)
- Comments (comment())
- Processing instructions (processing-instruction())

As such, when we list the results, we need a way to represent these different kind of nodes in a uniform manner.
This module aids by having objects to represent each visited XML node.
"""
from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2024-03-25'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from abc import ABCMeta
from dataclasses import dataclass
from lxml.etree import _Element, _ProcessingInstruction, _Comment


@dataclass(frozen=True, slots=True)
class XMLNode(metaclass=ABCMeta):
    """The base class for all XML nodes in this module.

    Args:
        xpath_location: The location of the provided element in XPath 3.1 notation, using the `BracedURILiteral`
            style for the qualified names.
    """
    xpath_location: str


@dataclass(frozen=True, slots=True)
class AttributeNode(XMLNode):
    """Represents attributes of XML elements.

    Args:
        name: the attribute name.
        value: a string value for the attribute
        parent: the parent element node
    """
    name: str
    value: str
    parent: _Element


@dataclass(frozen=True, slots=True)
class CommentNode(XMLNode):
    """Represents XML comments

    Args:
        element: the XML element
    """
    element: _Comment


@dataclass(frozen=True, slots=True)
class ProcessingInstructionNode(XMLNode):
    """Represents XML processing instructions nodes.

    Args:
        element: the wrapped Processing Instruction Element.
    """
    element: _ProcessingInstruction


@dataclass(frozen=True, slots=True)
class ElementNode(XMLNode):
    """Representation of an XML element

    Args:
        element: the wrapper XML element.
    """
    element: _Element



__author__ = 'Robbert Harms'
__date__ = '2023-02-16'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

import os
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import override

import elementpath

from lxml import etree
from lxml.etree import Element

from pyschematron.direct_mode.schematron.ast import Schema, Check, Assert, SchematronASTNode, Pattern, Rule, Report, \
    Variable, \
    Paragraph, ExtendsById, Extends, ExtendsExternal, ExternalRule, Query, QueryVariable, XMLVariable, \
    Namespace, Title, PatternParameter, ValueOf, Name, Phase, ActivePhase, Diagnostic, Diagnostics, Properties, \
    Property, XPathExpression
from pyschematron.direct_mode.schematron.parsers.xml.builders import ConcreteRuleBuilder, ExternalRuleBuilder, \
    AbstractRuleBuilder, ConcretePatternBuilder, SchemaBuilder, AbstractPatternBuilder, \
    InstancePatternBuilder, PhaseBuilder
from pyschematron.direct_mode.schematron.parsers.xml.utils import node_to_str, resolve_href, parse_attributes
from pyschematron.utils import load_xml_document


class ParserFactory(metaclass=ABCMeta):
    """Create a parser for parsing a specific XML element into an AST node.

    By using a factory method we allow subclassing the :class:`SchematronASTNode` elements and have a
    dedicated parser for such new subclasses without having to subclass a deep class hierarchy.

    For example, suppose you subclass the element "Assert" as "MyAssert" with some additional attributes.
    When parsing, you could reuse most of the parsing code by providing the factory with::

        {'assert': MyAssertParser()}

    Where `MyAssertParser` is a parser returning your subclassed "MyAssert" instead of "Assert".
    """

    @abstractmethod
    def get_parser(self, xml_tag: str) -> "ElementParser":
        """Get the parser for a specific Schematron XML tag.

        Args:
            xml_tag: the xml tag for which we want to get the parser

        Returns:
            A parser for the specific XML tag
        """


class DefaultParserFactory(ParserFactory):
    """The default parser factory.

    For each Schematron element, this returns the standard parser from this library.
    """

    def __init__(self):
        self.parsers = {
            'schema': SchemaParser(),
            'ns': NamespaceParser(),
            'phase': PhaseParser(),
            'active': ActivePhaseParser(),
            'pattern': PatternParser(),
            'rule': RuleParser(),
            'assert': AssertParser(),
            'report': ReportParser(),
            'extends': ExtendsParser(),
            'param': PatternParameterParser(),
            'diagnostics': DiagnosticsParser(),
            'diagnostic': DiagnosticParser(),
            'properties': PropertiesParser(),
            'property': PropertyParser(),
            'name': NameParser(),
            'value-of': ValueOfParser(),
            'let': VariableParser(),
            'p': ParagraphParser(),
            'title': TitleParser(),
            'include': IncludeParser(),
        }

    def get_parser(self, xml_tag: str) -> "ElementParser":
        return self.parsers[xml_tag]


class ParsingContext:

    def __init__(self, parser_factory: ParserFactory = None, base_path: Path = None):
        """Create the parser context we use while parsing the Schematron XML into the AST.

        Args:
            parser_factory: the factory we use to create sub parsers, defaults to :class:`DefaultParserFactory`.
            base_path: the base path to use for file inclusions, if not provided it is set to the current working
                directory.
        """
        self.parser_factory = parser_factory or DefaultParserFactory()
        self.base_path = base_path or os.getcwd()


class ElementParser(metaclass=ABCMeta):

    @abstractmethod
    def parse(self, element: Element, context: ParsingContext | None = None) -> SchematronASTNode:
        """Parse a piece of XML data into a SchematronASTNode.

        Args:
            element: the Schematron XML element we want to parse
            context: the context we use for parsing this Schematron XML element. If not set, we load a default parser.
        """

    @staticmethod
    def _parse_child_tags[T: Element](element: T, context: ParsingContext, xml_tag: str) -> list[T]:
        """Parse a sequence of XML elements with the same tag name.

        Args:
            element: the element which we search
            context: the parsing context
            xml_tag: the XML tag to search for
        """
        parser = context.parser_factory.get_parser(xml_tag)

        items = []
        for child in elementpath.select(element, xml_tag, namespaces={'': 'http://purl.oclc.org/dsdl/schematron'}):
            items.append(parser.parse(child, context))

        return items

    @staticmethod
    def get_rich_content(element: Element,
                         context: ParsingContext,
                         parse_special: bool = True,
                         remove_namespaces: bool = True) -> tuple[str | ValueOf | Name, ...]:
        """Get the rich content of the provided node.

        Args:
            element: the element for which to get the content.
            context: the parsing context
            parse_special: if set to True, we separate the strings and the special variables (ValueOf, Name).
                If set to false we return only string contents.
            remove_namespaces: if we want to remove the namespaces from the string rendered children

        Returns:
            A listing of text and special item nodes. All rich content like <emph> and <b> are rendered
            as string content. If parse special is True, Nodes of type <value-of> and <name> are parsed
            to ValueOf and Name.
        """
        content = [element.text]

        for child in element.getchildren():
            if parse_special and child.tag == '{http://purl.oclc.org/dsdl/schematron}value-of':
                parser = context.parser_factory.get_parser('value-of')
                content.append(parser.parse(child, context))
            elif parse_special and child.tag == '{http://purl.oclc.org/dsdl/schematron}name':
                parser = context.parser_factory.get_parser('name')
                content.append(parser.parse(child, context))
            else:
                content.append(node_to_str(child, remove_namespaces=remove_namespaces))
            content.append(child.tail)

        return tuple(el for el in content if el)


class SchemaParser(ElementParser):
    """Parse <schema> root tags"""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Schema:
        context = context or ParsingContext()

        builder = SchemaBuilder()
        builder.add_attributes(element.attrib)
        builder.add_patterns(self._parse_child_tags(element, context, 'pattern'))
        builder.add_namespaces(self._parse_child_tags(element, context, 'ns'))
        builder.add_phases(self._parse_child_tags(element, context, 'phase'))
        builder.add_diagnostics(self._parse_child_tags(element, context, 'diagnostics'))
        builder.add_properties(self._parse_child_tags(element, context, 'properties'))
        builder.add_paragraphs(self._parse_child_tags(element, context, 'p'))
        builder.add_variables(self._parse_child_tags(element, context, 'let'))

        if title_nodes := self._parse_child_tags(element, context, 'title'):
            builder.set_title(title_nodes[0])

        for include_node in self._parse_child_tags(element, context, 'include'):
            match include_node:
                case Pattern():
                    builder.add_patterns([include_node])
                case Namespace():
                    builder.add_namespaces([include_node])
                case Phase():
                    builder.add_namespaces([include_node])
                case Diagnostics():
                    builder.add_diagnostics([include_node])
                case Properties():
                    builder.add_properties([include_node])
                case Paragraph():
                    builder.add_paragraphs([include_node])
                case Variable():
                    builder.add_variables([include_node])
                case Title():
                    builder.set_title(include_node)

        return builder.build()


class NamespaceParser(ElementParser):
    """Parser for the `<ns>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Namespace:
        return Namespace(prefix=element.attrib['prefix'], uri=element.attrib['uri'])


class NameParser(ElementParser):
    """Parser for the `<name>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Name:
        if element.attrib.get('path'):
            return Name(Query(element.attrib['path']))
        return Name()


class ValueOfParser(ElementParser):
    """Parser for the `<value-of>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> ValueOf:
        return ValueOf(Query(element.attrib['select']))


class TitleParser(ElementParser):
    """Parser for the `<title>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Title:
        context = context or ParsingContext()
        return Title(content=''.join(self.get_rich_content(element, context)))


class PhaseParser(ElementParser):
    """Parser for the `<phase>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Phase:
        context = context or ParsingContext()
        builder = PhaseBuilder()

        builder.add_attributes(element.attrib)
        builder.add_active(self._parse_child_tags(element, context, 'active'))
        builder.add_variables(self._parse_child_tags(element, context, 'let'))
        builder.add_paragraphs(self._parse_child_tags(element, context, 'p'))

        for include_node in self._parse_child_tags(element, context, 'include'):
            match include_node:
                case ActivePhase():
                    builder.add_active([include_node])
                case Variable():
                    builder.add_variables([include_node])
                case Paragraph():
                    builder.add_paragraphs([include_node])

        return builder.build()


class ActivePhaseParser(ElementParser):
    """Parser for the `<active>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> ActivePhase:
        context = context or ParsingContext()
        content_parts = self.get_rich_content(element, context, parse_special=False)
        content = None
        if len(content_parts):
            content = ''.join(content_parts)

        return ActivePhase(pattern_id=element.attrib['pattern'], content=content)


class DiagnosticsParser(ElementParser):
    """Parser for the `<diagnostics>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Diagnostics:
        context = context or ParsingContext()
        diagnostics = []
        diagnostics.extend(self._parse_child_tags(element, context, 'diagnostic'))

        for include_node in self._parse_child_tags(element, context, 'include'):
            diagnostics.append(include_node)

        return Diagnostics(tuple(diagnostics))


class DiagnosticParser(ElementParser):
    """Parser for the `<diagnostic>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Diagnostic:
        context = context or ParsingContext()
        allowed_attributes = ['fpi', 'icon', 'id', 'role', 'see',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element.attrib, allowed_attributes, attribute_handlers)
        return Diagnostic(content=self.get_rich_content(element, context), **attributes)


class PropertiesParser(ElementParser):
    """Parser for the `<properties>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Properties:
        context = context or ParsingContext()
        properties = []
        properties.extend(self._parse_child_tags(element, context, 'property'))

        for include_node in self._parse_child_tags(element, context, 'include'):
            properties.append(include_node)

        return Properties(tuple(properties))


class PropertyParser(ElementParser):
    """Parser for the `<property>` tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Property:
        context = context or ParsingContext()
        attributes = parse_attributes(element.attrib, ['id', 'role', 'scheme'])
        return Property(content=self.get_rich_content(element, context), **attributes)


class PatternParser(ElementParser):
    """Parse <pattern> tags.

    If the pattern contains the abstract attribute set to True (`abstract="true"`), we load the
    pattern as an `AbstractPattern`. If the `is-a` attribute is present, we load the pattern as an `InstancePattern`.
    In other cases, we load the pattern as a `ConcretePattern`.
    """

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Pattern:
        context = context or ParsingContext()

        is_abstract = element.attrib.get('abstract', 'false') == 'true'
        is_instance = element.attrib.get('is-a')

        if is_abstract:
            builder = AbstractPatternBuilder()
        elif is_instance:
            builder = InstancePatternBuilder()
        else:
            builder = ConcretePatternBuilder()

        builder.add_attributes(element.attrib)
        builder.add_rules(self._parse_child_tags(element, context, 'rule'))
        builder.add_variables(self._parse_child_tags(element, context, 'let'))
        builder.add_paragraphs(self._parse_child_tags(element, context, 'p'))
        builder.add_parameters(self._parse_child_tags(element, context, 'param'))

        if title_nodes := self._parse_child_tags(element, context, 'title'):
            builder.set_title(title_nodes[0])

        for include_node in self._parse_child_tags(element, context, 'include'):
            match include_node:
                case Rule():
                    builder.add_rules([include_node])
                case Variable():
                    builder.add_variables([include_node])
                case Paragraph():
                    builder.add_paragraphs([include_node])
                case Title():
                    builder.set_title(include_node)

        return builder.build()


class PatternParameterParser(ElementParser):
    """Parse <param> tags used in instance patterns."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> PatternParameter:
        return PatternParameter(name=element.attrib['name'], value=element.attrib['value'])


class RuleParser(ElementParser):
    """Parse `<rule>` tags.

    Tags like `<rule abstract="true">` are parsed as `AbstractRule`.
    Tags like `<rule context="...">` are parsed as `ConcreteRule`.
    Tags missing both an abstract and context attribute are parsed as `ExternalRule`.
    """

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Rule:
        context = context or ParsingContext()

        is_abstract = element.attrib.get('abstract', 'false') == 'true'
        loaded_external = element.attrib.get('context') is None

        if is_abstract:
            builder = AbstractRuleBuilder()
        elif loaded_external:
            builder = ExternalRuleBuilder()
        else:
            builder = ConcreteRuleBuilder()

        builder.add_attributes(element.attrib)
        builder.add_checks(self._parse_child_tags(element, context, 'assert'))
        builder.add_checks(self._parse_child_tags(element, context, 'report'))
        builder.add_variables(self._parse_child_tags(element, context, 'let'))
        builder.add_paragraphs(self._parse_child_tags(element, context, 'p'))
        builder.add_extends(self._parse_child_tags(element, context, 'extends'))

        for include_node in self._parse_child_tags(element, context, 'include'):
            match include_node:
                case Check():
                    builder.add_checks([include_node])
                case Variable():
                    builder.add_variables([include_node])
                case Paragraph():
                    builder.add_paragraphs([include_node])
                case Extends():
                    builder.add_extends([include_node])

        return builder.build()


class IncludeParser(ElementParser):
    """Parse <include> tags.

    This parses the XML in the referenced file into a SchematronASTNode. Due to the generic nature of the include node
    this can be a node of any type. Note that Schematron includes do not support a multi-root XML document.
    """

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> SchematronASTNode:
        context = context or ParsingContext()
        file_path = resolve_href(element.attrib['href'], context.base_path)
        xml = load_xml_document(file_path).getroot()
        parser = context.parser_factory.get_parser(etree.QName(xml).localname)
        return parser.parse(xml, context)


class VariableParser(ElementParser):
    """Parse <let> tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Variable:
        context = context or ParsingContext()

        if 'value' in element.attrib:
            return QueryVariable(name=element.attrib['name'], value=Query(element.attrib['value']))
        else:
            content = ''.join(self.get_rich_content(element, context, parse_special=False, remove_namespaces=False))
            return XMLVariable(name=element.attrib['name'], value=content)


class ParagraphParser(ElementParser):
    """Parse <p> tags."""

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Paragraph:
        context = context or ParsingContext()

        allowed_attributes = ['icon', 'id', 'class',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            'class': lambda k, v: {'class_': v},
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element.attrib, allowed_attributes, attribute_handlers)
        content = ''.join(self.get_rich_content(element, context, parse_special=False))
        return Paragraph(content=content, **attributes)


class ExtendsParser(ElementParser):
    """Parse <extends> tags.

    If the extends tag points to another file, we will load the rule from that file and return it
    wrapped in an `ExtendsExternal`. If the `<extends>` points to a rule by ID, we return an `ExtendsById`.
    """

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Extends:
        context = context or ParsingContext()

        if 'rule' in element.attrib:
            return ExtendsById(element.attrib['rule'])

        file_path = resolve_href(element.attrib['href'], context.base_path)
        xml = load_xml_document(file_path).getroot()
        parser = context.parser_factory.get_parser('rule')
        rule = parser.parse(xml, context)

        if not isinstance(rule, ExternalRule):
            raise ValueError('The rule defined in the <extends> tag is invalid (contains context or is abstract).')

        return ExtendsExternal(rule, file_path)


class CheckParser(ElementParser):

    def __init__(self, type_instance: type[Check]):
        """Base parser for the <assert> and <report> checks.

        Args:
            type_instance: the type we are initializing, Assert or Report.
        """
        self.type_instance = type_instance

    @override
    def parse(self, element: Element, context: ParsingContext | None = None) -> Check:
        context = context or ParsingContext()

        allowed_attributes = ['test', 'diagnostics', 'properties', 'subject',
                              'id', 'role', 'flag', 'see', 'fpi', 'icon',
                              '{http://www.w3.org/XML/1998/namespace}lang',
                              '{http://www.w3.org/XML/1998/namespace}space']

        attribute_handlers = {
            'test': lambda k, v: {k: Query(v)},
            'diagnostics': lambda k, v: {k: tuple(v.split(' '))},
            'properties': lambda k, v: {k: tuple(v.split(' '))},
            'subject': lambda k, v: {k: XPathExpression(v)},
            '{http://www.w3.org/XML/1998/namespace}lang': lambda k, v: {'xml_lang': v},
            '{http://www.w3.org/XML/1998/namespace}space': lambda k, v: {'xml_space': v}
        }

        attributes = parse_attributes(element.attrib, allowed_attributes, attribute_handlers)
        return self.type_instance(content=self.get_rich_content(element, context), **attributes)


class AssertParser(CheckParser):

    def __init__(self):
        super().__init__(Assert)


class ReportParser(CheckParser):

    def __init__(self):
        super().__init__(Report)



"""This script shows how to use custom Python functions inside your Schematron schema's.

The general idea is that you either overwrite an existing query binding language, or define a new query binding
with your custom functions loaded. Your custom functions will then be attached to the query parser defined for that
specific query language.

As an example. Suppose you have a small custom function named `custom-func()`, and you want to use it in your
Schematron Schema. Your Schema is defined using `queryBinding="xpath31"` and you wish to extend this with your
custom function. For clarity, you want to call your new query binding language "xpath31-custom". In your Schematron
schema you then use `queryBinding="xpath31-custom"`, and in your queries you can use the `custom-func()`. For
PySchematron to know about this function, you must define it and add it to the library. This module shows how.

There are three ways of interacting with the PySchematron direct-mode validator. The most simple is by using the
functional interface defined in the main module. Second, you can use a generalized API which might be extended in the
future with an XSLT methodology. Finally, you can use the full-blown direct-mode classes and methods. The latter is the
most complicated but gives the most control. Either of these though enables adding custom functions.
"""

__author__ = 'Robbert Harms'
__date__ = '2024-04-03'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from pathlib import Path

from elementpath import ElementNode
from lxml import etree
from lxml.etree import _ElementTree

from pyschematron import DirectModeSchematronValidatorFactory, validate_document
from pyschematron.direct_mode.schematron.parsers.xml.parser import SchemaParser
from pyschematron.direct_mode.xml_validation.queries.factories import ExtendableQueryProcessorFactory
from pyschematron.direct_mode.xml_validation.queries.xpath import (XPathQueryProcessor, XPath31QueryParser,
                                                                   SimpleCustomXPathFunction)
from pyschematron.direct_mode.xml_validation.results.svrl_builder import DefaultSVRLReportBuilder
from pyschematron.direct_mode.xml_validation.validators import SimpleSchematronXMLValidator
from pyschematron.utils import load_xml_document


def get_example_schema() -> _ElementTree:
    """Get the example Schema for the examples.

    In this Schema, we defined a new query binding language `queryBinding="xpath31-custom"` which we will
    also need to add in PySchematron. Note also the use of `custom-func()`.

    Returns:
        The loaded Schema.
    """
    schematron = '''
    <schema xmlns="http://purl.oclc.org/dsdl/schematron"
            schemaVersion="iso"
            queryBinding="xpath31-custom"
            xml:lang="en"
            fpi="-//PYSCHEMATRON//DTD XML 1.0//EN">

        <ns prefix="c" uri="http://www.amazing-cargo.com/xml/data/2023"/>

        <pattern id="pa_check-banana">
            <p>Just a check on the banana's.</p>
            <rule subject="/c:cargo[1]" context="c:banana">
                <report test="@type = 'fruit'">Banana is a fruit <value-of select="custom-func(., 10)"/></report>
            </rule>
        </pattern>
    </schema>
    '''
    return load_xml_document(schematron)


def get_example_xml_document() -> _ElementTree:
    """Get the example XML document we wish to validate.

    This returns the XML document from the tests fixtures.

    Returns:
        The XML document we wish to validate.
    """
    return load_xml_document(Path('../tests/fixtures/full_example/cargo.xml'))


def custom_func(el: ElementNode, number: int) -> int:
    """An example of a custom function.

    It can have any number of inputs and outputs. This example takes an elementpath element as input and an integer.
    It returns the XPath node position times the provided number.
    """
    return el.position * number


def demo_functional_interface(xml_document: _ElementTree, schematron_xml: _ElementTree):
    """Showing how to add custom path functions using the functional interface.

    This uses the functional interface, the most simple method of interacting with PySchematron.

    Args:
        xml_document: the document we wish to validate
        schematron_xml: the Schematron Schema
    """
    custom_functions = {
        'query_binding': 'xpath31-custom',
        'base_query_binding': 'xpath31',
        'custom_query_functions': [SimpleCustomXPathFunction(custom_func, 'custom-func')]
    }

    result = validate_document(xml_document, schematron_xml, custom_functions=custom_functions)

    svrl = result.get_svrl()
    print(etree.tostring(svrl, pretty_print=True).decode('utf-8'))
    print(result.is_valid())


def demo_generic_api(xml_document: _ElementTree, schematron_xml: _ElementTree):
    """Showing how to add custom path functions using the general API.

    This uses the generic API which in the future might be extended using the XSLT method.

    Args:
        xml_document: the document we wish to validate
        schematron_xml: the Schematron Schema
    """
    validator_factory = DirectModeSchematronValidatorFactory(schematron_xml=schematron_xml)

    validator_factory.add_custom_functions('xpath31-custom',
                                           [SimpleCustomXPathFunction(custom_func, 'custom-func')], 'xpath31')

    validator = validator_factory.build()
    validation_result = validator.validate(xml_document)

    svrl = validation_result.get_svrl()
    print(etree.tostring(svrl, pretty_print=True).decode('utf-8'))
    print(validation_result.is_valid())


def demo_full_api(xml_document: _ElementTree, schematron_xml: _ElementTree):
    """Showing how to add custom path functions using the full direct-mode API.

    This is the most complex method, but shows how the direct-mode method operates.

    Args:
        xml_document: the document we wish to validate
        schematron_xml: the Schematron Schema

    """
    custom_xpath_function = SimpleCustomXPathFunction(custom_func, 'custom-func')

    custom_parser = XPath31QueryParser()
    custom_parser = custom_parser.with_custom_function(custom_xpath_function)

    custom_query_processor = XPathQueryProcessor(custom_parser)

    custom_processor_factory = ExtendableQueryProcessorFactory()
    custom_processor_factory.set_query_processor('xpath31-custom', custom_query_processor)

    schema = SchemaParser().parse(schematron_xml.getroot())

    validator = SimpleSchematronXMLValidator(schema, query_processor_factory=custom_processor_factory)
    validation_results = validator.validate_xml(xml_document)

    svrl = DefaultSVRLReportBuilder().create_svrl_xml(validation_results)

    print(etree.tostring(svrl, pretty_print=True).decode('utf-8'))
    print(validation_results.is_valid())


demo_functional_interface(get_example_xml_document(), get_example_schema())
demo_generic_api(get_example_xml_document(), get_example_schema())
demo_full_api(get_example_xml_document(), get_example_schema())




"""Implementation of the common API defined in the pyschematron package."""

__author__ = 'Robbert Harms'
__date__ = '2024-04-01'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from pathlib import Path

from lxml.etree import _ElementTree

from pyschematron.api import SchematronValidatorFactory, SchematronValidator, ValidationResult
from pyschematron.direct_mode.schematron.ast import Schema
from pyschematron.direct_mode.schematron.ast_visitors import ResolveExtendsVisitor, ResolveAbstractPatternsVisitor, \
    PhaseSelectionVisitor
from pyschematron.direct_mode.schematron.parsers.xml.parser import SchemaParser, ParsingContext
from pyschematron.direct_mode.xml_validation.queries.base import CustomQueryFunction
from pyschematron.direct_mode.xml_validation.queries.factories import ExtendableQueryProcessorFactory, \
    QueryProcessorFactory
from pyschematron.direct_mode.xml_validation.results.svrl_builder import DefaultSVRLReportBuilder
from pyschematron.direct_mode.xml_validation.results.validation_results import XMLDocumentValidationResult
from pyschematron.direct_mode.xml_validation.validators import SimpleSchematronXMLValidator
from pyschematron.utils import load_xml_document


class DirectModeSchematronValidatorFactory(SchematronValidatorFactory):

    def __init__(self,
                 schematron_xml: Path | _ElementTree | None = None,
                 phase: str | None = None,
                 schematron_base_path: Path | None = None):
        """Validator factory for direct mode Schematron validation.

        Args:
            schematron_xml: the Schematron Schema we want to evaluate. Can also be set using the class methods.
            phase: the phase we would like to evaluate. Can also be set using the class methods. If set to None
                we use the Schema's default phase.
            schematron_base_path: explicitly set the Schematron base path.
        """
        self._schematron_xml = schematron_xml
        self._phase = phase
        self._schematron_base_path = schematron_base_path
        self._custom_query_functions: dict[str, list[CustomQueryFunction]] = {}
        self._custom_query_bases: dict[str, str] = {}

    def set_schema(self, schematron_xml: Path | _ElementTree):
        self._schematron_xml = schematron_xml

    def set_phase(self, phase: str | None):
        self._phase = phase

    def set_base_path(self, schematron_base_path: Path):
        self._schematron_base_path = schematron_base_path

    def add_custom_functions(self,
                             query_binding: str,
                             custom_query_functions: list[CustomQueryFunction],
                             base_query_binding: str | None = None):
        if query_binding in self._custom_query_functions:
            self._custom_query_functions[query_binding] += custom_query_functions
        else:
            self._custom_query_functions[query_binding] = custom_query_functions

        if base_query_binding:
            self._custom_query_bases[query_binding] = base_query_binding

    def build(self) -> SchematronValidator:
        if isinstance(self._schematron_xml, Path):
            schematron = load_xml_document(self._schematron_xml)
        else:
            schematron = self._schematron_xml

        schematron_base_path = self._get_schematron_base_path()

        schematron_parser = SchemaParser()
        parsing_context = ParsingContext(base_path=schematron_base_path)
        schema = schematron_parser.parse(schematron.getroot(), parsing_context)

        query_processor_factory = self._get_query_processor_factory()

        return DirectModeSchematronValidator(schema, self._phase, schematron_base_path, query_processor_factory)

    def _get_query_processor_factory(self) -> QueryProcessorFactory | None:
        """Get the query processor factory we would like to use for the validation.

        Returns:
            The query processor to use, or None if defaults can be used.
        """
        if not len(self._custom_query_functions):
            return None

        custom_processor_factory = ExtendableQueryProcessorFactory()

        for query_binding, custom_functions in self._custom_query_functions.items():
            if query_binding in self._custom_query_bases:
                processor = custom_processor_factory.get_query_processor(self._custom_query_bases[query_binding])
            else:
                if custom_processor_factory.has_query_processor(query_binding):
                    processor = custom_processor_factory.get_query_processor(query_binding)
                else:
                    raise ValueError(f'No query binding base provided for adding '
                                     f'custom functions to query binding "{query_binding}"')

            for custom_function in custom_functions:
                processor = processor.with_custom_function(custom_function)

            custom_processor_factory.set_query_processor(query_binding, processor)
        return custom_processor_factory

    def _get_schematron_base_path(self) -> Path | None:
        """Get the base path we use for the Schematron parsing.

        If set explicitly, we return that. Else we try to infer it from the provided Schematron schema.

        Returns:
            The path we would like to use in the Schematron parsing.
        """
        if self._schematron_base_path is not None:
            return self._schematron_base_path

        if isinstance(self._schematron_xml, Path):
            return self._schematron_xml.parent

        if hasattr(self._schematron_xml, 'docinfo'):
            docinfo = self._schematron_xml.docinfo
            if hasattr(docinfo, 'URL') and docinfo.URL:
                return Path(docinfo.URL).parent

        return None


class DirectModeSchematronValidator(SchematronValidator):

    def __init__(self, schema: Schema,
                 phase: str | None,
                 base_path: Path | None,
                 query_processor_factory: QueryProcessorFactory | None = None):
        """Validator API implementation for the direct mode evaluation.

        Args:
            schema: the Schema we would like to use in the validation
            phase: the phase we would like to use
            base_path: the base path of the Schema, used for loading external Schema parts.
            query_processor_factory: optionally, specify the query processor factory to use.
        """
        self._schema = schema
        self._phase = phase
        self._base_path = base_path

        schema = ResolveExtendsVisitor(schema).apply(schema)
        schema = ResolveAbstractPatternsVisitor(schema).apply(schema)
        schema = PhaseSelectionVisitor(schema, phase).apply(schema)

        self._validator = SimpleSchematronXMLValidator(schema, phase, base_path,
                                                       query_processor_factory=query_processor_factory)

    def validate(self, xml_data: Path | _ElementTree) -> ValidationResult:
        if isinstance(xml_data, Path):
            xml_document = load_xml_document(xml_data)
        else:
            xml_document = xml_data

        validation_results = self._validator.validate_xml(xml_document)
        return DirectModeValidationResult(validation_results)


class DirectModeValidationResult(ValidationResult):

    def __init__(self, validation_results: XMLDocumentValidationResult):
        """Validation results from using the direct mode evaluation.

        Args:
            validation_results: the validation results from the direct mode evaluator.
        """
        self._validation_results = validation_results
        self._svrl_report = DefaultSVRLReportBuilder().create_svrl_xml(validation_results)
        self._is_valid = validation_results.is_valid()

    def get_svrl(self) -> _ElementTree:
        return self._svrl_report

    def is_valid(self) -> bool:
        return self._is_valid



__author__ = 'Robbert Harms'
__date__ = '2023-02-18'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

from pathlib import Path
from typing import Callable, Any

import lxml
from lxml import etree
from lxml.etree import _Element


def node_to_str(node: _Element, remove_namespaces: bool = True) -> str:
    """Convert an lxml node to string.

    This can be used to convert a node to string without namespaces.

    Args:
        node: the node to convert to string
        remove_namespaces: if we want to string the namespaces

    Returns:
        A string representation of the provided node.
    """
    tag_str = lxml.etree.tostring(node, with_tail=False, encoding='unicode')

    if remove_namespaces:
        new_root = etree.fromstring(tag_str)
        for elem in new_root.getiterator():
            elem.tag = etree.QName(elem).localname
        etree.cleanup_namespaces(new_root)

        return lxml.etree.tostring(new_root, encoding='unicode')
    return tag_str


def resolve_href(href: str, base_path: Path) -> Path:
    """Resolve a href attribute to a file on the filesystem.

    This can be used to resolve the `href` attributes of extend or include tags.

    Args:
        href: an absolute or relative path pointing to a file on the filesystem
        base_path: the base path to resolve relative paths.

    Returns:
        An absolute path to a file on the filesystem.
    """
    file_path = Path(href)
    if file_path.is_absolute():
        return file_path
    return (base_path / file_path).resolve()


def parse_attributes(attributes: dict[str, str],
                     allowed_attributes: list[str],
                     attribute_handlers: dict[str: Callable[[str, str], Any]] | None = None) -> dict[str, Any]:
    """Parse the attributes of the given element.

    By default, it returns all attributes as a string value. By using the attribute handlers it is possible
    to specify for each attribute how it is to be treated.

    Args:
        attributes: the attributes we wish to parse
        allowed_attributes: the set of allowed attributes, we will only parse and return the items in this list
        attribute_handlers: for each attribute name, a callback taking in the name and attribute value to return
            a new modified name and attribute value.

    Returns:
        For each allowed attribute the parsed values.
    """
    attribute_handlers = attribute_handlers or {}

    parsed_attributes = {}
    for item in allowed_attributes:
        if item in attributes:
            handler = attribute_handlers.get(item, lambda k, v: {k: v})
            parsed_attributes.update(handler(item, attributes[item]))
    return parsed_attributes



__author__ = 'Robbert Harms'
__date__ = '2024-03-17'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from abc import ABCMeta

from pyschematron.direct_mode.lib.ast import GenericASTVisitor
from pyschematron.direct_mode.svrl.ast import SVRLNode


class SVRLASTVisitor(GenericASTVisitor[SVRLNode], metaclass=ABCMeta):
    """Visitor implementation for the SVRL nodes."""
    ...



from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2023-02-21'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

import dataclasses
import inspect
from abc import ABCMeta, abstractmethod
from io import StringIO
from pathlib import PosixPath, Path
from typing import Callable, Any, Mapping, Iterable

from ruyaml import YAML, BaseRepresenter, BaseConstructor, Node

import pyschematron.direct_mode.schematron.ast
from pyschematron.direct_mode.schematron.ast import SchematronASTNode


class ASTYamlConverter(metaclass=ABCMeta):

    @abstractmethod
    def load(self, stream: Path | Any) -> SchematronASTNode:
        """Load a provided stream into a Schematron AST node.

        Args:
            stream: the stream with Yaml data to load, either a path to a file, or some sort of (binary) string.

        Returns:
            The loaded Schematron AST node.
        """

    @abstractmethod
    def dump(self, data: SchematronASTNode, stream: Path | Any | None = None) -> str | None:
        """Dump the provided data into YAML format.

        Args:
            data: the Schematron AST node to dump
            stream: the stream to dump the data to, either a file path, or some sort of string buffer.
                This is optional, if not provided we return a string.

        Returns:
            If no stream was provided, we return the output as a string. Else we dump the output to the provided stream.
        """


class RuyamlASTYamlConverter(ASTYamlConverter):

    def __init__(self):
        """Basic AST to YAML (and back) converter.

        This works together with the ruyaml library to dump and load AST nodes to and from yaml.
        """
        self._codec = _ASTYamlCodec()

    def load(self, stream: Path | Any) -> SchematronASTNode:
        return self._codec.load(stream)

    def dump(self, data: SchematronASTNode, stream: Path | Any | None = None) -> str | None:
        if stream is None:
            with StringIO() as dumped:
                self._codec.dump(data, dumped)
                return dumped.getvalue()
        else:
            self._codec.dump(data, stream)


class _ASTYamlCodec(YAML):

    def __init__(self, *args, **kwargs):
        """Specialized RuYAML loader and dumper for :class:`SchematronASTNode` instances.

        For loading the YAML, the ruyaml library uses shared mutable lists to construct the class hierarchy.
        Since we are creating immutable AST nodes, we need a builder pattern to construct the final classes.
        During processing of the YAML, the :class:`YamlRepresenter` create :class:`SchematronASTNodeBuilder` nodes.
        As a final step, the build method of these builders is called to construct the final AST nodes.
        """
        super().__init__(*args, typ='safe', **kwargs)
        self._add_ast_representers()

    def load(self, stream: Path | Any) -> Any:
        loaded_objects = super().load(stream)
        if isinstance(loaded_objects, SchematronASTNodeBuilder):
            return loaded_objects.build()
        return loaded_objects

    def _add_ast_representers(self):
        """Add YAML representers (loading and dumping) for all relevant types in `pyschematron.ast`.

        This modifies the current instance by adding representer objects.
        """
        representers = self._get_ast_node_representers()
        representers.append(PathRepresenter())

        for representer in representers:
            self.representer.add_representer(representer.element_class, representer.get_dumping_function())
            self.constructor.add_constructor(representer.yaml_tag, representer.get_loading_function())

    def _get_ast_node_representers(self) -> list["YamlRepresenter"]:
        """Get representers for all the :class:`SchematronASTNode` nodes.

        Returns:
            A list of YAML representers for the AST nodes.
        """
        ast_nodes = self._list_representable_ast_nodes()
        return [self._get_representer(ast_node) for ast_node in ast_nodes]

    def _list_representable_ast_nodes(self) -> list[type[SchematronASTNode]]:
        """List the representable AST nodes.

        Returns:
            Get a list of AST nodes we can YAML represent.
        """
        def filter_function(el):
            return inspect.isclass(el) and issubclass(el, SchematronASTNode) and el is not SchematronASTNode
        return [el[1] for el in inspect.getmembers(pyschematron.direct_mode.schematron.ast, filter_function)]

    def _get_representer(self, ast_node: type[SchematronASTNode]) -> "YamlRepresenter":
        """Get a representor for the indicated node type.

        Args:
            ast_node: the type of AST node for which we want a representer.

        Returns:
            A representer for this specific node type.
        """
        return GenericASTNodeYamlRepresenter(ast_node)


class YamlRepresenter(metaclass=ABCMeta):
    """Specialized class for YAML representing elements and contents of :class:`SchematronASTNode`.

    This works in conjunction with the Ruyaml library.
    """

    @property
    @abstractmethod
    def element_class(self) -> Any:
        """Get the type of element this representer represents."""

    @property
    @abstractmethod
    def yaml_tag(self) -> str:
        """Get the YAML tag for the class representation."""

    @abstractmethod
    def get_dumping_function(self) -> Callable[[BaseRepresenter, Any], Callable]:
        """Get the dumping function `to_yaml` we can use to dump the item to YAML.

        Returns:
            A function to create the yaml representation.
        """

    @abstractmethod
    def get_loading_function(self) -> Callable[[BaseConstructor, Node], Any]:
        """Get the loading function we can use to load the YAML node into a `SchematronASTNode`.

        Returns:
            A function to load the yaml representation.
        """


class PathRepresenter(YamlRepresenter):

    @property
    def element_class(self) -> Any:
        return PosixPath

    @property
    def yaml_tag(self) -> str:
        return f'!Path'

    def get_dumping_function(self) -> Callable[[BaseRepresenter, Any], Node]:
        def to_yaml(representer: BaseRepresenter, node: Any) -> Node:
            return representer.represent_scalar(self.yaml_tag, str(node))
        return to_yaml

    def get_loading_function(self) -> Callable[[BaseConstructor, Node], Any]:
        def from_yaml(constructor: BaseConstructor, node: Node) -> Any:
            return self.element_class(node.value)
        return from_yaml


class ASTNodeYamlRepresenter(YamlRepresenter):

    def get_dumping_function(self) -> Callable[[BaseRepresenter, Any], Node]:
        def to_yaml(representer: BaseRepresenter, node: SchematronASTNode) -> Node:
            return self._to_yaml(representer, node)
        return to_yaml

    def get_loading_function(self) -> Callable[[BaseConstructor, Node], Any]:
        def from_yaml(constructor: BaseConstructor, node: Node) -> SchematronASTNodeBuilder:
            return self._from_yaml(constructor, node)
        return from_yaml

    @abstractmethod
    def _to_yaml(self, representer: BaseRepresenter, node: SchematronASTNode) -> Node:
        """The YAML dumping function.

        This is the actual function used to dump a specific Schematron node to YAML.

        Returns:
            The ruyaml Node used to dump to YAML.
        """

    @abstractmethod
    def _from_yaml(self, constructor: BaseConstructor, node: Node) -> SchematronASTNodeBuilder:
        """The function to load a ruyaml YAML Node into a Schematron AST Node.

        Returns:
            A builder for the Schematron node.
        """


class GenericASTNodeYamlRepresenter(ASTNodeYamlRepresenter):

    def __init__(self, node_type: type[SchematronASTNode]):
        """Create a basic representer for :class:`SchematronASTNode`.

        When loading a YAML, this class returns :class:`SchematronASTNodeBuilder` nodes.

        Args:
            node_type: the type of node we are representing
        """
        self._node_type = node_type

    @property
    def element_class(self) -> type[SchematronASTNode]:
        return self._node_type

    @property
    def yaml_tag(self) -> str:
        return f'!{self._node_type.__name__}'

    def _to_yaml(self, representer: BaseRepresenter, node: SchematronASTNode) -> Node:
        """Internal function called by the forwarding function in `get_dumping_function`"""
        init_names = [f.name for f in dataclasses.fields(self.element_class) if f.init]

        node_data = {}
        for init_name in init_names:
            if value := getattr(node, init_name):
                node_data[init_name] = value

        return representer.represent_mapping(self.yaml_tag, node_data)

    def _from_yaml(self, constructor: BaseConstructor, node: Node) -> SchematronASTNodeBuilder:
        """Internal function called by the forwarding function in `get_loading_function`"""
        values = constructor.construct_mapping(node)
        return DictionaryNodeBuilder(self._node_type, values)


class SchematronASTNodeBuilder(metaclass=ABCMeta):
    """Builder pattern for delayed construction of Schematron nodes."""

    @abstractmethod
    def build(self) -> SchematronASTNode:
        """Build a Schematron node based on the information in this builder.

        Returns:
            The constructed schematron element.
        """


class DictionaryNodeBuilder(SchematronASTNodeBuilder):

    def __init__(self, node_type: type[SchematronASTNode], init_values: dict):
        """Construct a Schematron AST node using a dictionary containing the init values.

        During the build phase, this builder will transform the provided init values according to these rules:
        1. if a node contains a builder, build it
        2. if a node contains a list, modify it to become a tuple

        Args:
            node_type: the type of node we will build
            init_values: the init values to pass to the constructor.
        """
        self._node_type = node_type
        self._init_values = init_values

    def build(self) -> SchematronASTNode:
        def _expand_value(value):
            if isinstance(value, SchematronASTNodeBuilder):
                return value.build()
            elif isinstance(value, str):
                return value
            elif isinstance(value, Mapping):
                return {k: _expand_value(v) for k, v in value.items()}
            elif isinstance(value, Iterable):
                return tuple(_expand_value(el) for el in value)
            else:
                return value

        final_inits = {}
        for key, value in self._init_values.items():
            final_inits[key] = _expand_value(value)

        return self._node_type(**final_inits)



__author__ = 'Robbert Harms'
__date__ = '2023-03-06'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

from typing import Any, Mapping, Iterable, Literal, override

from abc import ABCMeta

from pyschematron.direct_mode.lib.ast import GenericASTVisitor
from pyschematron.direct_mode.schematron.ast import SchematronASTNode, Schema, ConcretePattern, Rule, ExtendsExternal, \
    ExternalRule, ConcreteRule, ExtendsById, AbstractRule, AbstractPattern, InstancePattern, Pattern, Phase
from pyschematron.direct_mode.schematron.utils import macro_expand


class SchematronASTVisitor(GenericASTVisitor[SchematronASTNode], metaclass=ABCMeta):
    """Visitor pattern for the Schematron AST nodes."""


class FindIdVisitor(SchematronASTVisitor):

    def __init__(self, id_ref: str):
        """A visitor which finds a node with the given ID.

        Args:
            id_ref: the id we would like to find in the visited nodes
        """
        super().__init__()
        self._id_ref = id_ref

    @override
    def visit(self, ast_node: SchematronASTNode) -> Any:
        if hasattr(ast_node, 'id') and getattr(ast_node, 'id') == self._id_ref:
            return ast_node

        for child in ast_node.get_children():
            if found_node := self.visit(child):
                return found_node


class GetIDMappingVisitor(SchematronASTVisitor):

    def __init__(self):
        """A visitor which maps all nodes with an id to their id."""
        super().__init__()
        self._result = {}

    @override
    def visit(self, ast_node: SchematronASTNode) -> Any:
        self._result |= self._visit(ast_node)
        return self._result

    def _visit(self, ast_node: SchematronASTNode) -> dict[str, SchematronASTNode]:
        for child in ast_node.get_children():
            child.accept_visitor(self)

        if hasattr(ast_node, 'id'):
            if (node_id := getattr(ast_node, 'id')) is not None:
                return {node_id: ast_node}
        return {}


class GetNodesOfTypeVisitor(SchematronASTVisitor):

    def __init__(self, types: type[SchematronASTNode] | tuple[type[SchematronASTNode], ...]):
        """A visitor which checks each node for their type against the type(s) provided

        Args:
            types: a single type or a tuple of types we check each node against.
        """
        super().__init__()
        self._types = types
        self._result = []

    @override
    def visit(self, ast_node: SchematronASTNode) -> Any:
        for child in ast_node.get_children():
            child.accept_visitor(self)

        if isinstance(ast_node, self._types):
            self._result.append(ast_node)

        return self._result


class ResolveExtendsVisitor(SchematronASTVisitor):

    def __init__(self, schema: Schema):
        """Simplify an AST Schema by inlining all the extends in the rules.

        This visitor inlines the variables and checks of each of the extended rules.
        `AbstractRule` and `ExternalRule` items are deleted after inlining.

        Args:
            schema: the full Schema as input to lookup all the rules by ID.
        """
        super().__init__()
        self._schema = schema

    @override
    def visit(self, ast_node: SchematronASTNode) -> SchematronASTNode:
        match ast_node:
            case Schema():
                return self._process_schema(ast_node)
            case ConcretePattern() | AbstractPattern():
                return self._process_pattern(ast_node)
            case Rule():
                return self._process_rule(ast_node)
            case ExtendsExternal():
                return self._process_extends_external(ast_node)
            case ExtendsById():
                return self._process_extends_by_id(ast_node)
            case _:
                return ast_node

    def _process_schema(self, schema: Schema) -> Schema:
        """Process a Schema by processing all the patterns.

        Args:
            schema: the schema to process

        Returns:
            A processed schema
        """
        patterns = []
        for pattern in schema.patterns:
            patterns.append(ResolveExtendsVisitor(self._schema).apply(pattern))
        return schema.with_updated(patterns=tuple(patterns))

    def _process_pattern(self, pattern: ConcretePattern | AbstractPattern) -> ConcretePattern | AbstractPattern:
        """Process a pattern by processing all the rules.

        Args:
            pattern: the pattern to process

        Returns:
            the processed pattern
        """
        rules = []
        for rule in pattern.rules:
            processed_rule = ResolveExtendsVisitor(self._schema).apply(rule)
            if isinstance(processed_rule, ConcreteRule):
                rules.append(processed_rule)
        return pattern.with_updated(rules=tuple(rules))

    def _process_rule(self, rule: Rule) -> Rule:
        """Process a rule by inlining all the extends.

        Args:
            rule: the rule we wish to process

        Returns:
            A new rule with all the extends loaded and added to the checks.
        """
        extra_checks = []
        extra_variables = []
        for extends in rule.extends:
            extended_rule = ResolveExtendsVisitor(self._schema).apply(extends)
            extra_checks.extend(extended_rule.checks)
            extra_variables.extend(extended_rule.variables)

        checks = tuple(extra_checks) + rule.checks
        variables = tuple(extra_variables) + rule.variables
        return rule.with_updated(checks=checks, variables=variables, extends=tuple())

    def _process_extends_by_id(self, extends: ExtendsById) -> AbstractRule:
        """Process an extends which points to an abstract rule.

        Args:
            extends: the extends node we are processing

        Returns:
            The abstract rule this extends points to.
        """
        abstract_rule = FindIdVisitor(extends.id_ref).apply(self._schema)
        if abstract_rule is None:
            raise ValueError(f'Can\'t find the abstract rule with id "{extends.id_ref}"')
        return ResolveExtendsVisitor(self._schema).apply(abstract_rule)

    def _process_extends_external(self, extends: ExtendsExternal) -> ExternalRule:
        """Process an external extend by returning the loaded rule.

        Args:
            extends: the extends node we are processing

        Returns:
            The loaded external rule
        """
        return ResolveExtendsVisitor(self._schema).apply(extends.rule)


class ResolveAbstractPatternsVisitor(SchematronASTVisitor):

    def __init__(self, schema: Schema):
        """Simplify an AST Schema by expanding all the instance-of patterns.

        This visitor substitutes the abstract patterns with each of the instance-of patterns.
        All abstract patterns are deleted from the AST after replacement.

        Args:
            schema: the full Schema as input to lookup all the rules by ID.
        """
        super().__init__()
        self._schema = schema

    @override
    def visit(self, ast_node: SchematronASTNode) -> SchematronASTNode:
        match ast_node:
            case Schema():
                return self._process_schema(ast_node)
            case InstancePattern():
                return self._process_instance_pattern(ast_node)
            case _:
                return ast_node

    def _process_schema(self, schema: Schema) -> Schema:
        """Process a Schema by processing all the patterns.

        Args:
            schema: the schema to process

        Returns:
            A processed schema
        """
        patterns = []
        for pattern in schema.patterns:
            new_pattern = ResolveAbstractPatternsVisitor(self._schema).apply(pattern)
            if isinstance(new_pattern, ConcretePattern):
                patterns.append(new_pattern)

        return schema.with_updated(patterns=tuple(patterns))

    def _process_instance_pattern(self, instance_pattern: InstancePattern) -> ConcretePattern:
        """Process an instance-of pattern by expanding it with an abstract pattern.

        Args:
            instance_pattern: the instance-of pattern to process

        Returns:
            the processed pattern as a concrete pattern
        """
        abstract_pattern = FindIdVisitor(instance_pattern.abstract_id_ref).apply(self._schema)
        if abstract_pattern is None:
            raise ValueError(f'Can\'t find the abstract pattern with id "{instance_pattern.abstract_id_ref}"')

        macro_expansions = {f'${param.name}': param.value for param in instance_pattern.params}
        macro_expand_visitor = MacroExpandVisitor(macro_expansions)

        macro_expanded_pattern = macro_expand_visitor.apply(abstract_pattern)
        return macro_expanded_pattern.with_updated(id=instance_pattern.id)


class MacroExpandVisitor(SchematronASTVisitor):

    def __init__(self, macro_expansions: dict[str, str]):
        """Macro expand an abstract pattern.

        If the input is an AbstractPattern we return a ConcretePattern.
        In all other cases we return a node of the same type but with macro expanded elements.

        Args:
            A mapping of macro expansions to apply.
        """
        super().__init__()
        self._macro_expansions = macro_expansions

    @override
    def visit(self, ast_node: SchematronASTNode) -> SchematronASTNode:
        if isinstance(ast_node, AbstractPattern):
            expanded_pattern = self._visit_generic_node(ast_node)
            return ConcretePattern(**expanded_pattern.get_init_values())

        return self._visit_generic_node(ast_node)

    def _visit_generic_node[T: SchematronASTNode](self, ast_node: T) -> T:
        """Visit a generic node and do macro expansion.

        Args:
            ast_node: the node we are visiting and expanding

        Returns:
            A node of the same type but with macro expanded items.
        """
        sub_visitor = MacroExpandVisitor(self._macro_expansions)
        init_values = ast_node.get_init_values()

        def _expand_value(value):
            if isinstance(value, str):
                return macro_expand(value, self._macro_expansions)
            elif isinstance(value, SchematronASTNode):
                return sub_visitor.apply(value)
            elif isinstance(value, Mapping):
                return {k: _expand_value(v) for k, v in value.items()}
            elif isinstance(value, Iterable):
                return tuple(_expand_value(el) for el in value)
            else:
                return value

        updated_items = {}
        for key, value in init_values.items():
            updated_items[key] = _expand_value(value)

        return ast_node.with_updated(**updated_items)


class PhaseSelectionVisitor(SchematronASTVisitor):

    def __init__(self, schema: Schema, phase: str | Literal['#ALL', '#DEFAULT'] | None = None):
        """Reduce an AST to only those patterns and phases referenced by a specific phase.

        This visitor only works on concrete Schema AST trees, we assume all abstract rules and patterns to be resolved.

        The output limits the `patterns` in the AST to only those selected by the phase.
        It will also limit the `phases` to the active phase, or to an empty list if no phase was specified.

        Args:
            schema: the full Schema as input to lookup all the rules by ID.
            phase: the phase we want to select, can be an IDREF of a phase node, the literal `#ALL` for all patterns,
                or `#DEFAULT` for the `defaultPhase` attribute of the Schematron. The default value is `#DEFAULT`,
                it is overwritten by the attribute `defaultPhase`, which again can be overwritten by the phase
                here specified.
        """
        super().__init__()
        self._schema = schema
        self._phase = phase
        self._phase_node = self._get_phase_node(schema, phase)

        self._active_pattern_ids = None
        if self._phase_node:
            self._active_pattern_ids = [active_phase.pattern_id for active_phase in self._phase_node.active]

    @override
    def visit(self, ast_node: SchematronASTNode) -> SchematronASTNode | bool:
        match ast_node:
            case Schema():
                return self._process_schema(ast_node)
            case Pattern():
                return self._process_pattern(ast_node)
            case Phase():
                return self._process_phase(ast_node)
            case _:
                return ast_node

    def _process_schema(self, schema: Schema) -> Schema:
        """Process a Schema by reducing the patterns and phases to the specified set.

        Args:
            schema: the schema to process

        Returns:
            A processed schema
        """
        patterns = tuple(pattern for pattern in schema.patterns if self.apply(pattern))
        phases = tuple(phase for phase in schema.phases if self.apply(phase))
        return schema.with_updated(patterns=patterns, phases=phases)

    def _process_pattern(self, pattern: Pattern) -> bool:
        """Process a pattern by verifying if it is in the current phase.

        Args:
            pattern: the pattern to process

        Returns:
            A boolean indicating if this pattern is in the current phase or not.
        """
        if not isinstance(pattern, ConcretePattern):
            raise ValueError('This visitor can only deal with concrete patterns.')

        return self._active_pattern_ids is None or pattern.id in self._active_pattern_ids

    def _process_phase(self, phase: Phase) -> bool:
        """Process a phase node by verifying if it is in the current phase.

        Args:
            phase: the phase to process

        Returns:
            A boolean indicating if this phase is in the current phase or not
        """
        return self._phase_node is None or phase.id == self._phase_node.id

    def _get_phase_node(self, schema: Schema, phase: str | Literal['#ALL', '#DEFAULT'] | None = None) -> Phase | None:
        """Get the phase node associated with the elected phase, or None if None found.

        Args:
            schema: the schema we want to search
            phase: the chosen phase.

        Returns:
            The AST phase node, or None if not applicable / not found.
        """
        if phase is None:
            phase = '#DEFAULT'

        if phase == '#ALL':
            return None

        if phase == '#DEFAULT':
            phase = schema.default_phase

        if isinstance(phase, str):
            phase_node = FindIdVisitor(phase).apply(self._schema)

            if phase_node is None:
                raise ValueError(f'Can not find the phase "{phase}".')
            return phase_node

        return None



"""The abstract syntax tree for representing an SVRL report.

This is primarily used to write out an SVRL report after Schematron validation of an XML.
"""
from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2024-03-11'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from dataclasses import dataclass
from typing import Literal

from lxml.etree import _Element

from pyschematron.direct_mode.lib.ast import GenericASTNode


@dataclass(slots=True, frozen=True)
class SVRLNode(GenericASTNode):
    """Base class for the Schematron Validation Report Language (SVRL) nodes."""


@dataclass(slots=True, frozen=True)
class SchematronOutput(SVRLNode):
    """Representation of the `<schematron-output>` SVRL node.

    Args:
        texts: zero or more text nodes containing some text about the schema and/or validation.
        ns_prefix_in_attribute_values: namespace and prefix declarations.
        validation_events: schematron validation events, in order of proceeding
        phase: the Schematron phase this SVRL is a result of
        schema_version: copy of the Schematron's schemaVersion attribute
        title: some title for this validation report.
    """
    texts: tuple[Text, ...] = tuple()
    ns_prefix_in_attribute_values: tuple[NSPrefixInAttributeValues, ...] = tuple()
    validation_events: tuple[ValidationEvent, ...] = tuple()
    metadata: MetaData | None = None
    phase: str | None = None
    schema_version: str | None = None
    title: str | None = None


@dataclass(slots=True, frozen=True)
class MetaData(SVRLNode):
    """Metadata for this SVRL report.

    A non-standard defined node containing metadata. We use it to add metadata about PySchematron.

    Args:
        xml_elements: listing of the XML elements contained in this metadata
        namespaces: the namespaces to be used in the attributes of this metadata node.
    """
    xml_elements: tuple[_Element, ...] = tuple()
    namespaces: tuple[Namespace, ...] = tuple()

    @dataclass(slots=True, frozen=True)
    class MetaDataNode:
        """Base class for metadata nodes."""

    @dataclass(slots=True, frozen=True)
    class Namespace(MetaDataNode):
        """Representation of a namespace attribute in the metadata node.

        Args:
            prefix: the prefix
            uri: the namespace's URI
        """
        prefix: str
        uri: str


@dataclass(slots=True, frozen=True)
class NSPrefixInAttributeValues(SVRLNode):
    """Namespace declaration, representation of the SVRL `<ns-prefix-in-attribute-values>` node.

    Args:
        prefix: the prefix to use for this namespace
        uri: the namespace's URI
    """
    prefix: str
    uri: str


@dataclass(slots=True, frozen=True)
class ValidationEvent(SVRLNode):
    """Base class for the validation events.

    An SVRL is a flat representation of the patterns, rules and assertions / reports visited during validation.
    To represent these in a class hierarchy we group these as validation events.
    """


@dataclass(slots=True, frozen=True)
class ActivePattern(ValidationEvent):
    """Representation of the `<active-pattern>` SVRL node.

    Args:
        documents: list of URIs of datatype `xs:anyURI`, pointing to the documents processed.
        id: the identifier of this pattern, typically a copy of the Schematron pattern id.
        name: some name for this pattern, up to the implementation.
        role: some role indicator for this pattern, up to the implementation.
    """
    documents: tuple[str, ...] | None = None
    id: str | None = None
    name: str | None = None
    role: str | None = None


@dataclass(slots=True, frozen=True)
class FiredRule(ValidationEvent):
    """Representation of the `<fired-rule>` SVRL node.

    Args:
        context: a copy of the context of the Schematron rule element
        document: Reference of the document to which this rule was defined.
        flag: a flag that was set to true when this rule fired, typically a copy of the flag of the Schematron rule.
        id: the identifier of this rule, typically a copy of the Schematron rule id.
        name: some name for this rule, up to the implementation.
        role: the role for this rule, typically a copy of the role of the rule element.
    """
    context: SchematronQuery
    document: str | None = None
    flag: str | None = None
    id: str | None = None
    name: str | None = None
    role: str | None = None


@dataclass(slots=True, frozen=True)
class SuppressedRule(ValidationEvent):
    """Representation of the `<suppressed-rule>` SVRL node.

    This node type is officially not in the standard, but it is added by some packages, and so do we.

    Args:
        context: a copy of the context of the Schematron rule element
        id: the identifier of this rule, typically a copy of the Schematron rule id.
    """
    context: SchematronQuery
    id: str | None = None


@dataclass(slots=True, frozen=True)
class CheckResult(ValidationEvent):
    """Base class for the `<failed-assert>` and `<successful-report>` SVRL nodes.

    Args:
        text: result description of this check.
        location: the location of this failed assert as an XPath expression
        test: the test expression for this assert, copied from the Schematron assert node.
        diagnostic_references: listing of the diagnostic references by this check
        property_references: properties referenced by this check
        subject_location: the location referenced by the subject of either the check or the parent rule.
        flag: a flag that was set to true when this assertion fired, typically a copy of the Schematron's flag rule.
        id: the identifier of this rule, typically a copy of the Schematron assert id.
        role: the role for this assert, typically a copy of the role of the assert element.
    """
    text: Text
    location: XPathExpression
    test: SchematronQuery
    diagnostic_references: tuple[DiagnosticReference, ...] = tuple()
    property_references: tuple[PropertyReference, ...] = tuple()
    subject_location: XPathExpression | None = None
    flag: str | None = None
    id: str | None = None
    role: str | None = None


@dataclass(slots=True, frozen=True)
class FailedAssert(CheckResult):
    """Representation of the `<failed-assert>` SVRL node."""


@dataclass(slots=True, frozen=True)
class SuccessfulReport(CheckResult):
    """Representation of the `<successful-report>` SVRL node."""


@dataclass(slots=True, frozen=True)
class DiagnosticReference(SVRLNode):
    """Representation of the `<diagnostic-reference>` node.

    These nodes reference a diagnostic connected to an assert/report node.

    Args:
        text: resulting text
        diagnostic: identifier of this diagnostic, copied from the diagnostic element's id attribute
    """
    text: Text
    diagnostic: str


@dataclass(slots=True, frozen=True)
class PropertyReference(SVRLNode):
    """Representation of the `<property-reference>` node.

    These nodes reference a property connected to an assert/report node.

    Args:
        text: resulting text
        property: identifier of this property
        role: the role attribute for this property, copied from the properties' role attribute
        scheme: the scheme attribute for this property, copied from the properties' scheme attribute
    """
    text: Text
    property: str
    role: str | None = None
    scheme: str | None = None


@dataclass(slots=True, frozen=True)
class XPathExpression(SVRLNode):
    """Representation of an XPath expression  used in the SVRL nodes."""
    expression: str


@dataclass(slots=True, frozen=True)
class SchematronQuery(SVRLNode):
    """Representation of a Schematron Query, as used in the SVRL nodes."""
    query: str


@dataclass(slots=True, frozen=True)
class Text(SVRLNode):
    """Representation of a `<text>` tag.

    Although the attributes `class` and `id` are not specified in the SVRL specification, we add them nonetheless since
    they can be forwarded from Schematron nodes.

    Args:
        content: the text content of this text element, all loaded as one string
        fpi: formal public identifier, may be copied from the relevant Schematron FPI attribute.
        icon: the icon attribute
        see: A URI pointing to some external information of this element.
        class_: some class attribute
        id: unique identifier
        xml_lang: the default natural language for this node
        xml_space: defines how whitespace must be handled for this element.
    """
    content: str
    fpi: str | None = None
    icon: str | None = None
    see: str | None = None
    class_: str | None = None
    id: str | None = None
    xml_lang: str | None = None
    xml_space: Literal['default', 'preserve'] | None = None



__author__ = 'Robbert Harms'
__date__ = '2020-02-04'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@laltoida.com'


from importlib import metadata
from importlib.metadata import PackageNotFoundError
from pathlib import Path

import tomllib

try:
    __version__ = metadata.version('pyschematron')
except PackageNotFoundError:
    with open(Path(__file__).parent.parent / 'pyproject.toml', 'rb') as f:
        pyproject = tomllib.load(f)
        __version__ = pyproject['project']['version']



__author__ = 'Robbert Harms'
__date__ = '2023-02-17'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'


from io import BytesIO, IOBase
from pathlib import Path
from typing import BinaryIO, Union

from lxml import etree
from lxml.etree import _ElementTree


def load_xml_document(xml_data: Union[bytes, str, Path, IOBase, BinaryIO]) -> _ElementTree:
    """Load an XML document from a polymorphic source.

    Args:
        xml_data: the XML data to load. Can be loaded from a string, file, or byte-like object.

    Returns:
        The document node of the loaded XML.
    """
    match xml_data:
        case IOBase():
            parser = etree.XMLParser(ns_clean=True)
            return etree.parse(xml_data, parser)
        case bytes():
            return load_xml_document(BytesIO(xml_data))
        case str():
            return load_xml_document(BytesIO(xml_data.encode('utf-8')))
        case Path():
            with open(xml_data, 'rb') as f:
                return load_xml_document(f)



"""This script demonstrates how to load a Schematron Schema in the PySchematron direct-mode.

This example shows the use of the direct-mode Abstract Syntax Tree (AST) for PySchematron Schemas. By loading Schematron
schema's in the AST, you can inspect the Schema using Python functionality.

Please note that only Schematron specific XML nodes are loaded from the Schematron Schema. Custom nodes are not loaded.
You can however augment the AST and the parser with your own nodes. This is however not demonstrated (here).
"""

__author__ = 'Robbert Harms'
__date__ = '2023-02-21'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

from pathlib import Path

from pyschematron.direct_mode.schematron.ast_visitors import ResolveExtendsVisitor, ResolveAbstractPatternsVisitor, \
    PhaseSelectionVisitor
from pyschematron.direct_mode.schematron.ast_yaml import RuyamlASTYamlConverter
from pyschematron.direct_mode.schematron.parsers.xml.parser import ParsingContext, SchemaParser
from pyschematron.utils import load_xml_document


schematron_path = Path('../tests/fixtures/full_example/schema.sch')
schematron_xml = load_xml_document(schematron_path)
parsing_context = ParsingContext(base_path=schematron_path.parent)

# Parse the Schema
schematron_parser = SchemaParser()
schema = schematron_parser.parse(schematron_xml.getroot(), parsing_context)

# Shows the use of the visitor pattern to modify the Schema
schema = ResolveExtendsVisitor(schema).apply(schema)
schema = ResolveAbstractPatternsVisitor(schema).apply(schema)
schema = PhaseSelectionVisitor(schema, '#ALL').apply(schema)

# Experimental, YAML conversion.
yaml_converter = RuyamlASTYamlConverter()
yaml_shema = yaml_converter.dump(schema)
round_trip = yaml_converter.load(yaml_shema)

print(yaml_shema)
print(round_trip == schema)



from __future__ import annotations

"""The XPath query bindings.

This uses the adapter pattern around the elementpath library.

The elementpath library is used for the parsing, context and evaluation of all XPath related queries.
"""

__author__ = 'Robbert Harms'
__date__ = '2023-03-24'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'GPL v3'

from typing import Any, Type, Self, override
from abc import ABCMeta

from elementpath import XPathToken, XPath1Parser, XPath2Parser, XPathContext
from elementpath.xpath3 import XPath3Parser
from elementpath.xpath31 import XPath31Parser
from elementpath.xpath_context import ItemArgType
from elementpath.tree_builders import RootArgType

from pyschematron.direct_mode.xml_validation.queries.base import QueryParser, Query, EvaluationContext, \
    SimpleQueryProcessor, CustomQueryFunction, SimpleCustomQueryFunction
from pyschematron.direct_mode.xml_validation.queries.exceptions import MissingRootNodeError


class XPathQueryProcessor(SimpleQueryProcessor):

    def __init__(self, query_parser: XPathQueryParser, evaluation_context: EvaluationContext = None):
        """Query processor for XPath queries.

        This is simply a wrapper around the simple query processor, with as default evaluation context the XPath
        evaluation context.

        Args:
            query_parser: the (XPath) query processor to use
            evaluation_context: the evaluation context, defaults to the XPath evaluation context
        """
        super().__init__(query_parser, evaluation_context or XPathEvaluationContext())


class XPathQueryParser(QueryParser, metaclass=ABCMeta):
    """Wrapper around the query parser to indicate specialization for XPath query parsers."""


class ElementPathXPathQueryParser(XPathQueryParser, metaclass=ABCMeta):

    def __init__(self,
                 parser_type: Type[XPath1Parser | XPath2Parser | XPath3Parser | XPath31Parser],
                 namespaces: dict[str, str] | None = None,
                 custom_functions: list[CustomXPathFunction] | None = None):
        """Base class for XPath parsers wrapping the `elementpath` library.

        Args:
            parser_type: the type of XPath parser from the elementpath library we are wrapping
            namespaces: namespaces to use during parsing
            custom_functions: the list of custom functions to load
        """
        self._parser_type = parser_type
        self._namespaces = namespaces or {}
        self._custom_functions = custom_functions or []
        self._parser = self._get_elementpath_parser()

    @override
    def parse(self, source: str) -> Query:
        xpath_token = self._parser.parse(source)
        return XPathQuery(xpath_token)

    def _get_elementpath_parser(self) -> XPath1Parser | XPath2Parser | XPath3Parser | XPath31Parser:
        """Get an elementpath parser using the defined namespaces and custom functions.

        Returns:
            An elementpath parser instance to use during parsing.
        """
        parser = self._parser_type(namespaces=self._namespaces)
        for custom_function in self._custom_functions:
            parser.external_function(custom_function.callback, custom_function.name, custom_function.prefix)
        return parser


class CustomXPathFunction(CustomQueryFunction, metaclass=ABCMeta):
    """Wrapper around the custom query functions to specify functions for use in XPath query parsers."""


class SimpleCustomXPathFunction(CustomXPathFunction, SimpleCustomQueryFunction):
    """Simple definition of an XPath custom function."""


class XPathEvaluationContext(EvaluationContext):

    def __init__(self,
                 root: RootArgType | None = None,
                 namespaces: dict[str, str] | None = None,
                 item: ItemArgType | None = None,
                 variables: dict[str, Any] | None = None):
        super().__init__()
        self._context_variables = {
            'root': root,
            'namespaces': namespaces or {},
            'item': item,
            'variables': variables or {}
        }

        self._xpath_context = None
        if root is not None:
            self._xpath_context = XPathContext(**self._context_variables)

    def get_xpath_context(self) -> XPathContext | None:
        """Get the XPath context we can use for evaluation of a query.

        If no root node is set yet, we return None. Else, we return an XPathContext from the elementpath library.

        Returns:
             The XPath context if a root node is set, else None.

        Raises:
            MissingRootNodeError: if the XPath node could not be
        """
        if self._xpath_context is None:
            raise MissingRootNodeError('Missing root node in XPath context, please set a root node first.')
        return self._xpath_context

    @override
    def with_context_item(self, xml_item: ItemArgType) -> Self:
        if xml_item is self._context_variables['item']:
            return self
        return self._get_updated({'item': xml_item})

    @override
    def with_xml_root(self, xml_root: RootArgType) -> Self:
        if xml_root is self._context_variables['root']:
            return self
        return self._get_updated({'root': xml_root})

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return self._get_updated({'namespaces': namespaces})

    @override
    def with_variables(self, variables: dict[str, Any], overwrite: bool = False) -> Self:
        if overwrite:
            return self._get_updated({'variables': variables})
        else:
            return self._get_updated({'variables': self._context_variables['variables'] | variables})

    @override
    def get_xml_root(self) -> RootArgType | None:
        return self._context_variables['root']

    @override
    def get_context_item(self) -> ItemArgType | None:
        return self._context_variables['item']

    def _get_updated(self, updates: dict[str, Any]) -> Self:
        kwargs = self._context_variables.copy()
        kwargs.update(updates)
        return type(self)(**kwargs)


class XPathQuery(Query):

    def __init__(self, xpath_token: XPathToken):
        """Representation of an XPath query.

        This uses the elementpath library for representing the XPath expressions.

        Args:
            xpath_token: the parsed XPath expression
        """
        self._xpath_token = xpath_token

    @override
    def evaluate(self, context: XPathEvaluationContext | None = None) -> Any:
        xpath_context = None
        if context:
            xpath_context = context.get_xpath_context()

        return self._xpath_token.evaluate(xpath_context)


class XPath1QueryParser(ElementPathXPathQueryParser):

    def __init__(self, namespaces: dict[str, str] | None = None):
        """Query parser for XPath 1.0 expressions.

        This uses the XPath 1.0 parser of the `elementpath` library.

        Args:
             namespaces: a dictionary with namespaces to use while parsing.
        """
        super().__init__(XPath1Parser, namespaces=namespaces)

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return type(self)(self._namespaces | namespaces)

    @override
    def with_custom_function(self, custom_function: CustomXPathFunction) -> Self:
        raise ValueError('Custom functions are not supported for XPath1 parsers.')


class XPath2QueryParser(ElementPathXPathQueryParser):

    def __init__(self,
                 namespaces: dict[str, str] | None = None,
                 custom_functions: list[CustomXPathFunction] | None = None):
        """Query parser for XPath 2.0 expressions.

        This uses the XPath 2.0 parser of the `elementpath` library.

        Args:
             namespaces: a dictionary with namespaces to use while parsing.
             custom_functions: the list of custom functions to load
        """
        super().__init__(XPath2Parser, namespaces=namespaces, custom_functions=custom_functions)

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return type(self)(self._namespaces | namespaces,
                          custom_functions=self._custom_functions)

    @override
    def with_custom_function(self, custom_function: CustomXPathFunction) -> Self:
        return type(self)(namespaces=self._namespaces,
                          custom_functions=self._custom_functions + [custom_function])


class XPath3QueryParser(ElementPathXPathQueryParser):

    def __init__(self,
                 namespaces: dict[str, str] | None = None,
                 custom_functions: list[CustomXPathFunction] | None = None):
        """Query parser for XPath 3.0 expressions.

        This uses the XPath 3.0 parser of the `elementpath` library.

        Args:
             namespaces: a dictionary with namespaces to use while parsing.
             custom_functions: the list of custom functions to load
        """
        super().__init__(XPath3Parser, namespaces=namespaces, custom_functions=custom_functions)

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return type(self)(self._namespaces | namespaces,
                          custom_functions=self._custom_functions)

    @override
    def with_custom_function(self, custom_function: CustomXPathFunction) -> Self:
        return type(self)(namespaces=self._namespaces,
                          custom_functions=self._custom_functions + [custom_function])


class XPath31QueryParser(ElementPathXPathQueryParser):

    def __init__(self,
                 namespaces: dict[str, str] | None = None,
                 custom_functions: list[CustomXPathFunction] | None = None):
        """Query parser for XPath 3.1 expressions.

        This uses the XPath 3.1 parser of the `elementpath` library.

        Args:
             namespaces: a dictionary with namespaces to use while parsing.
             custom_functions: the list of custom functions to load
        """
        super().__init__(XPath31Parser, namespaces=namespaces, custom_functions=custom_functions)

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return type(self)(self._namespaces | namespaces,
                          custom_functions=self._custom_functions)

    @override
    def with_custom_function(self, custom_function: CustomXPathFunction) -> Self:
        return type(self)(namespaces=self._namespaces,
                          custom_functions=self._custom_functions + [custom_function])



__author__ = 'Robbert Harms'
__date__ = '2024-04-06'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from pathlib import Path

import typer

from pyschematron import validate_documents, validate_document

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)


@app.command(help='Validate one or more documents using PySchematron.')
def validate(xml_documents: list[Path] = typer.Argument(help='One or more documents to validate.',
                                                        exists=True, file_okay=True, dir_okay=False, resolve_path=True),
             schema: Path = typer.Argument(help='The Schematron Schema to use for the validation.',
                                           exists=True, file_okay=True, dir_okay=False, resolve_path=True),
             phase: str = typer.Option('#DEFAULT', '--phase', '-p', help='The Schematron phase to use.'),
             svrl_out: Path = typer.Option(None, '--svrl-out',
                                           help='The file to write the SVRL to. '
                                                'For multiple documents we append the XML document name to this name.',
                                           file_okay=True, dir_okay=False, writable=True)):

    if svrl_out:
        svrl_out.parent.mkdir(parents=True, exist_ok=True)

    if len(xml_documents) == 1:
        result = validate_document(xml_documents[0], schema, phase=phase)

        print(xml_documents[0], 'VALID' if result.is_valid() else 'INVALID')

        if svrl_out:
            result.get_svrl().write(str(svrl_out), pretty_print=True, xml_declaration=True, encoding="utf-8")
    else:
        results = validate_documents(xml_documents, schema, phase=phase)

        for filename, result in zip(xml_documents, results):
            print(filename, 'VALID' if result.is_valid() else 'INVALID')

            if svrl_out:
                out_fname = svrl_out.with_stem(svrl_out.stem + '_' + Path(filename).stem)
                result.get_svrl().write(str(out_fname), pretty_print=True, xml_declaration=True, encoding="utf-8")



from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2023-02-18'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

import numbers
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Literal, Any

from elementpath import TextNode, XPathNode
from elementpath.tree_builders import get_node_tree
from elementpath.xpath31 import XPath31Parser
from elementpath.xpath_context import ItemArgType, XPathContext
from lxml.etree import ElementTree, tostring, _Element

from pyschematron.direct_mode.schematron.ast import (Schema, ConcretePattern, ConcreteRule, Variable,
                                                     XMLVariable, QueryVariable, Assert, Report, ValueOf, Name,
                                                     XPathExpression, Property, Diagnostic)
from pyschematron.direct_mode.schematron.ast_visitors import (ResolveExtendsVisitor, ResolveAbstractPatternsVisitor,
                                                              PhaseSelectionVisitor, FindIdVisitor)

from pyschematron.direct_mode.xml_validation.queries.base import EvaluationContext, Query, QueryParser, \
    CachingQueryParser
from pyschematron.direct_mode.xml_validation.queries.factories import DefaultQueryProcessorFactory, QueryProcessorFactory
from pyschematron.direct_mode.xml_validation.results.validation_results import (CheckResult, PatternResult,
                                                                                FullNodeResult,
                                                                                XMLDocumentValidationResult,
                                                                                SuppressedRuleResult, FiredRuleResult,
                                                                                SkippedRuleResult, XMLInformation,
                                                                                SchemaInformation, PropertyResult,
                                                                                DiagnosticResult)
from pyschematron.direct_mode.xml_validation.results.xml_nodes import XMLNode, ProcessingInstructionNode, ElementNode, \
    CommentNode, AttributeNode


class SchematronXMLValidator(metaclass=ABCMeta):
    """Base class for Schematron XML validators.

    Schematron XML validators are used to validate an XML and return a validation report.
    """

    @abstractmethod
    def validate_xml(self, xml_document: ElementTree) -> XMLDocumentValidationResult:
        """Validate the provided XML using this Schematron XML validator.

        Args:
            xml_document: the XML document to validate in an lxml `elementtree` object

        Returns:
            An XML document result containing the result of processing all patterns over all nodes.
        """


class SimpleSchematronXMLValidator(SchematronXMLValidator):

    def __init__(self,
                 schema: Schema,
                 phase: str | Literal['#ALL', '#DEFAULT'] | None = None,
                 schematron_base_path: Path | None = None,
                 query_processor_factory: QueryProcessorFactory = None):
        """XML validation using a Schematron Schema.

        Validation is done into two phases. In the first phase we create a hierarchy of validator objects. These
        validator objects are prepared with a query parser for parsing the queries and are aware of the Schematron
        schema. During evaluation (e.g. after calling the method `validate_xml`), they are provided with the XML
        node under investigation and are called with a context to be used for evaluating the queries.

        Args:
            schema: the Schema AST node. We will make the AST concrete if not yet done so.
            phase: the phase we want to evaluate. If None, we evaluate the default phase.
            schematron_base_path: the base path from which we loaded the Schematron file, provided for context.
            query_processor_factory: the query processor factory we would like to use to load the query processor
                from the Schema. By providing this we allow injection of custom query processors.
        """
        self._phase = phase
        self._schema = self._reduce_schema_to_phase(schema, self._phase)
        self._schematron_base_path = schematron_base_path
        self._query_processor_factory = query_processor_factory or DefaultQueryProcessorFactory()

        self._query_processor = self._query_processor_factory.get_schema_query_processor(schema)
        self._query_parser = CachingQueryParser(self._query_processor.get_query_parser())
        self._variable_evaluators = _get_variable_evaluators(self._schema.variables, self._query_parser)
        self._pattern_validators = self._get_pattern_validators()

    def validate_xml(self, xml_document: ElementTree) -> XMLDocumentValidationResult:
        xml_tree = get_node_tree(root=xml_document)

        root_context = self._query_processor.get_evaluation_context().with_xml_root(xml_tree)
        context = _get_context_with_variables(self._variable_evaluators, root_context)

        node_results = []
        for node in xml_tree.iter_lazy():
            if not node.parent:
                continue

            if not isinstance(node, TextNode):
                if node_result := self._validate_node(node, context):
                    node_results.append(node_result)

        xml_information = XMLInformation(xml_document)
        schema_information = SchemaInformation(self._schema, self._phase, self._schematron_base_path)
        return XMLDocumentValidationResult(xml_information, schema_information, tuple(node_results))

    def _validate_node(self, node: ItemArgType, evaluation_context: EvaluationContext) -> FullNodeResult:
        """Validate the indicated XML node.

        Args:
            node: the XML node to validate using all the patterns in the Schematron
            evaluation_context: the context we use to parse the queries.

        Returns:
            Either None if no applicable results, or else the result of visiting the node.
        """
        pattern_results = []
        for pattern_validator in self._pattern_validators:
            pattern_results.append(pattern_validator.validate(node, evaluation_context))

        return FullNodeResult(_to_result_node(node), evaluation_context, tuple(pattern_results))

    def _reduce_schema_to_phase(self, schema: Schema, phase: str | Literal['#ALL', '#DEFAULT'] | None = None) -> Schema:
        """Reduce an AST to only those patterns and phases referenced by a specific phase.

        This will first resolve all AST abstractions and afterward reduce the Schema to the phase selected.

        Reducing a Schema means selecting only those patterns and phases which fall within the selected phase.

        Args:
            schema: the Schema to resolve and reduce
            phase: the selected phase

        Returns:
            The resolved and reduced Schema
        """
        schema = self._resolve_ast_abstractions(schema)
        return PhaseSelectionVisitor(schema, phase=phase).apply(schema)

    @staticmethod
    def _resolve_ast_abstractions(schema: Schema) -> Schema:
        """Resolve all abstractions in a provided AST Schema node and return a Schema node without abstractions.

        This resolves the `extends` in the Rules, and resolves the abstract patterns.

        Args:
             schema: the Schema to resolve all abstractions in

        Returns:
            The Schema with the abstractions resolved.
        """
        schema = ResolveExtendsVisitor(schema).apply(schema)
        schema = ResolveAbstractPatternsVisitor(schema).apply(schema)
        return schema

    def _get_pattern_validators(self) -> list[_PatternValidator]:
        """Parse the patterns in the AST Schema and return a list of validators.

        Returns:
            A list of pattern validators.
        """
        pattern_validators = []
        for pattern in self._schema.patterns:
            if not isinstance(pattern, ConcretePattern):
                raise ValueError(f'Schema not concrete, ConcretePattern expected, {type(pattern)} received.')

            if len(pattern.rules):
                pattern_validators.append(_PatternValidator(self._schema, self._query_parser, pattern))

        return pattern_validators


class _PatternValidator:

    def __init__(self, schema: Schema, query_parser: QueryParser, pattern: ConcretePattern):
        """The Pattern validator validates an XML node against a pattern.

        Args:
            schema: the entire Schematron schema we are applying
            pattern: the Schematron pattern we would like to apply in the validation step.
            query_parser: the parser we can use to parse queries in the pattern.
        """
        self._schema = schema
        self._query_parser = query_parser
        self._pattern = pattern
        self._variable_evaluators = _get_variable_evaluators(self._pattern.variables, self._query_parser)
        self._rule_validators: list[_RuleValidator] = self._get_rule_validators()

    def validate(self, xml_node: ItemArgType, evaluation_context: EvaluationContext) -> PatternResult:
        """Validate the XML node using the encapsulated pattern.

        For building the report, this will apply all the rules in the pattern against the provided XML node.
        According to Schematron, the first rule whose context matches the XML node will be used as the rule for that
        node within this pattern. Nevertheless, we apply all rules in order to build a comprehensive report.

        Args:
            xml_node: the node we are validating
            evaluation_context: the context to use for validation

        Returns:
            A report with the results of validating this pattern.
        """
        context = _get_context_with_variables(self._variable_evaluators, evaluation_context)

        nmr_fired_rules = 0
        rule_results = []
        for rule_validator in self._rule_validators:
            rule_result = rule_validator.validate(xml_node, context)

            if rule_result.is_fired():
                if nmr_fired_rules == 0:
                    rule_results.append(rule_result)
                else:
                    rule_results.append(SuppressedRuleResult.from_fired_rule_result(rule_result))
                nmr_fired_rules += 1
            else:
                rule_results.append(rule_result)

        return PatternResult(_to_result_node(xml_node), evaluation_context, self._pattern, tuple(rule_results))

    def _get_rule_validators(self) -> list[_RuleValidator]:
        """Initialize the rule validators.

        Returns:
            The list of rule validators obtained from the pattern
        """
        rule_validators = []
        for rule in self._pattern.rules:
            if not isinstance(rule, ConcreteRule):
                raise ValueError(f'Schema not concrete, ConcreteRule expected, {type(rule)} received.')
            rule_validators.append(_RuleValidator(self._schema, self._query_parser, rule))

        return rule_validators


class _RuleValidator:

    def __init__(self, schema: Schema, query_parser: QueryParser, rule: ConcreteRule):
        """The Rule validator validates an XML node against a single rule.

        Args:
            schema: the entire Schematron schema we are applying
            rule: the Schematron rule we would like to apply in the validation step.
            query_parser: the parser we can use to parse queries in the pattern.
        """
        self._schema = schema
        self._query_parser = query_parser
        self._rule = rule
        self._variable_evaluators = _get_variable_evaluators(self._rule.variables, self._query_parser)
        self._check_validators = self._get_check_validators()
        self._context_query = self._query_parser.parse(self._rule.context.query)

    def validate(self,
                 xml_node: ItemArgType,
                 evaluation_context: EvaluationContext) -> SkippedRuleResult | FiredRuleResult:
        """Validate the XML node using the encapsulated rule.

        This first checks if the context of the rule matches the XML node. If it matches we apply all the assertions
        and report checks in the rule. If it does not match we return the negative match as a result.

        Args:
            xml_node: the node we are validating
            evaluation_context: the context to use for validation

        Returns:
            A report with the results of validating this rule.
        """
        if not self._node_matches_context(xml_node, evaluation_context):
            return SkippedRuleResult(_to_result_node(xml_node), evaluation_context, self._rule)

        context = _get_context_with_variables(self._variable_evaluators, evaluation_context)
        check_results = []
        for check_validator in self._check_validators:
            check_results.append(check_validator.validate(xml_node, context))

        subject_node = get_subject_node(self._rule.subject, self._query_parser, context)

        return FiredRuleResult(_to_result_node(xml_node), evaluation_context, self._rule, check_results, subject_node)

    def _node_matches_context(self, xml_node: ItemArgType, evaluation_context: EvaluationContext) -> bool:
        """Check if the node we are investigating matches the context of a rule.

        We perform the check by querying the parent of the node for a match on the context of the rule.
        There may be multiple matches, as such we check if the node is within the matches. If so, the node
        must match the context query of the Schematron rule.

        Args:
            xml_node: the node we are investigating
            evaluation_context: the evaluation context. We will relativize this to the parent of the XML node.

        Returns:
            True if the node matches the rule, false otherwise.
        """
        parent_context = evaluation_context.with_context_item(xml_node.parent)

        if matches := self._context_query.evaluate(parent_context):
            if xml_node in matches:
                return True
        return False

    def _get_check_validators(self) -> list[_CheckValidator]:
        """Initialize the assert and report validators.

        Returns:
            The list of check validators obtained from the rule.
        """
        check_validators = []
        for check in self._rule.checks:
            if not isinstance(check, (Assert, Report)):
                raise ValueError(f'Each check should either be an Assert or Report node, {type(check)} received.')
            check_validators.append(_CheckValidator(self._schema, self._query_parser, check))

        return check_validators


class _CheckValidator:

    def __init__(self, schema: Schema, query_parser: QueryParser, check: Assert | Report):
        """Validate an XML node using the indicated Schematron assert or report.

        Args:
            schema: the entire Schematron schema we are applying
            query_parser: the parser we can use to parse queries in the check.
            check: the check node to apply to the XML node
        """
        self._schema = schema
        self._query_parser = query_parser
        self._check = check
        self._check_query = self._query_parser.parse(check.test.query)
        self._rich_text_content_evaluator = _RichTextContentEvaluator(self._check.content, query_parser)

        self._property_evaluators = []
        if self._check.properties:
            for property_id in self._check.properties:
                property = FindIdVisitor(property_id).apply(schema)
                self._property_evaluators.append(_PropertyEvaluator(property, query_parser))

        self._diagnostic_evaluators = []
        if self._check.diagnostics:
            for diagnostic_id in self._check.diagnostics:
                diagnostic = FindIdVisitor(diagnostic_id).apply(schema)
                self._diagnostic_evaluators.append(_DiagnosticEvaluator(diagnostic, query_parser))

    def validate(self, xml_node: ItemArgType, evaluation_context: EvaluationContext) -> CheckResult:
        """Validate the XML node using the encapsulated check (report or assert).

        Args:
            xml_node: the node we are validating
            evaluation_context: the context to use for validation

        Returns:
            A report with the results of validating this check.
        """
        context = evaluation_context.with_context_item(xml_node)

        check_result = self._get_check_result_value(self._check_query.evaluate(context))
        text = self._rich_text_content_evaluator.evaluate(context)
        subject_node = get_subject_node(self._check.subject, self._query_parser, context)

        property_results = self._process_properties(context)
        diagnostic_results = self._process_diagnostics(context)

        return CheckResult(_to_result_node(xml_node), evaluation_context, self._check, check_result, text,
                           subject_node, tuple(property_results), tuple(diagnostic_results))

    def _get_check_result_value(self, query_result: Any) -> bool:
        """Translate the result of the query to a boolean.

        The query result may return a mixed bag of lists, strings and other values which should translate to a single
        value: True or False. The following rules apply. Any string or numeric value is translated to a boolean value
        using Python's "bool()" command. Lists or tuples are evaluated recursively and if all terminal values yield
        true, the result is true, false otherwise.

        Args:
            query_result: the result of the query, may be a mixed bag

        Returns:
            The boolean truth value of the query result
        """
        if isinstance(query_result, (bool, str, numbers.Number)):
            return bool(query_result)
        elif isinstance(query_result, XPathNode):
            return True
        elif isinstance(query_result, (list, tuple)):
            if not len(query_result):
                return False
            return all(map(self._get_check_result_value, query_result))

    def _process_properties(self, context: EvaluationContext) -> list[PropertyResult]:
        """Process the optional properties of the check and return a list of property results.

        Args:
            context: the context used to translate the ValueOf and Name elements

        Returns:
            The processed properties.
        """
        property_results = []
        for property_evaluator in self._property_evaluators:
            property_results.append(property_evaluator.evaluate(context))
        return property_results

    def _process_diagnostics(self, context: EvaluationContext) -> list[DiagnosticResult]:
        """Process the optional diagnostics of the check and return a list of diagnostic results.

        Args:
            context: the context used to translate the ValueOf and Name elements

        Returns:
            The processed diagnostics
        """
        diagnostic_results = []
        for diagnostic_evaluator in self._diagnostic_evaluators:
            diagnostic_results.append(diagnostic_evaluator.evaluate(context))
        return diagnostic_results


class _DelayedQueryEvaluation(metaclass=ABCMeta):
    """Representation of objects containing Schematron queries which may be lazily evaluated."""

    @abstractmethod
    def evaluate(self, context: EvaluationContext) -> Any:
        """Evaluate the parsed variable using the provided context.

        Args:
            context: the context we need to evaluate a query.
        """


class _StringVariableEvaluator(_DelayedQueryEvaluation):

    def __init__(self, value: str):
        """Representation of a string variable.

        Args:
            value: the string value
        """
        self._value = value

    def evaluate(self, context: EvaluationContext) -> str:
        return self._value


class _QueryVariableEvaluator(_DelayedQueryEvaluation):

    def __init__(self, value: Query):
        """Representation of a Query variable.

        Args:
            value: the query ready to be evaluated.
        """
        self._value = value

    def evaluate(self, context: EvaluationContext) -> Any:
        return self._value.evaluate(context)


class _PropertyEvaluator(_DelayedQueryEvaluation):

    def __init__(self, property: Property, query_parser: QueryParser):
        """Representation of a Schematron property with possible queries already parsed.

        Args:
            property: the Schematron property we are pre-parsing
            query_parser: the parser we can use to parse the queries inside the rich text content.
        """
        self._property = property
        self._content_evaluator = None
        if property.content:
            self._content_evaluator = _RichTextContentEvaluator(property.content, query_parser)

    def evaluate(self, context: EvaluationContext) -> PropertyResult:
        text = ''
        if self._content_evaluator:
            text = self._content_evaluator.evaluate(context)
        return PropertyResult(text, self._property.id, self._property.role, self._property.scheme)


class _DiagnosticEvaluator(_DelayedQueryEvaluation):

    def __init__(self, diagnostic: Diagnostic, query_parser: QueryParser):
        """Representation of a Schematron diagnostic with possible queries already parsed.

        Args:
            diagnostic: the Schematron diagnostic we are pre-parsing
            query_parser: the parser we can use to parse the queries inside the rich text content.
        """
        self._diagnostic = diagnostic
        self._content_evaluator = None
        if diagnostic.content:
            self._content_evaluator = _RichTextContentEvaluator(diagnostic.content, query_parser)

    def evaluate(self, context: EvaluationContext) -> DiagnosticResult:
        text = ''
        if self._content_evaluator:
            text = self._content_evaluator.evaluate(context)
        return DiagnosticResult(text, self._diagnostic.id, xml_lang=self._diagnostic.xml_lang,
                                xml_space=self._diagnostic.xml_space)


class _RichTextContentEvaluator(_DelayedQueryEvaluation):

    def __init__(self, content: tuple[str | ValueOf | Name, ...], query_parser: QueryParser):
        """Process rich text content of a Schematron node.

        Some Schematron nodes may contain rich text with various layout XML elements and two Schematron elements,
        ValueOf and Name. These need to be parsed and rendered correctly.

        Args:
            content: the rich text content
            query_parser: the parser we may use to parse queries.
        """
        self._content_elements = []

        for text_element in content:
            if isinstance(text_element, str):
                self._content_elements.append(text_element)
            elif isinstance(text_element, ValueOf):
                self._content_elements.append(query_parser.parse(text_element.select.query))
            elif isinstance(text_element, Name):
                if text_element.path:
                    self._content_elements.append(query_parser.parse(text_element.path.query.rstrip('/') + '/name()'))
                else:
                    self._content_elements.append(query_parser.parse('./name()'))

    def evaluate(self, context: EvaluationContext) -> str:
        """Evaluate the rich text content and return it as a string.

        Args:
            context: the current evaluation content

        Returns:
            The rendered content as a string.
        """
        processed_text = []
        for content_element in self._content_elements:
            if isinstance(content_element, str):
                processed_text.append(content_element)
            elif isinstance(content_element, Query):
                query_result = content_element.evaluate(context)
                if isinstance(query_result, list):
                    for query_el in query_result:
                        if isinstance(query_el, XPathNode):
                            if isinstance(query_el.value, _Element):
                                processed_text.append(tostring(query_el.value).decode())
                            else:
                                processed_text.append(query_el.value)
                        else:
                            processed_text.append(str(query_el))
                else:
                    processed_text.append(str(query_result))

        return (''.join(processed_text)).strip()


def _get_variable_evaluators(variables: tuple[Variable, ...], parser: QueryParser) -> dict[str, _DelayedQueryEvaluation]:
    """Parse a list of Schematron variables into evaluator objects.

    Args:
        variables: the AST variables we would like to parse.
        parser: the parser to use for parsing the queries.

    Returns:
        The parsed variables ready for evaluation.
    """
    results = {}
    for variable in variables:
        if isinstance(variable, QueryVariable):
            results[variable.name] = _QueryVariableEvaluator(parser.parse(variable.value.query))
        elif isinstance(variable, XMLVariable):
            results[variable.name] = _StringVariableEvaluator(variable.value)
    return results


def _get_context_with_variables(variable_evaluators: dict[str, _DelayedQueryEvaluation], context: EvaluationContext):
    """Update the provided context with the evaluated variables.

    This will evaluate the provided variable evaluators with the provided context and add the result of the
    variable evaluators to an updated context. Updating happens recursively since variables may refer to each other.
    It returns the updated context.

    Args:
        variable_evaluators: the evaluators for the various variables, indexed by name
        context: the context used to evaluate the variables. Will be updated with said variables and returned.

    Returns:
        The context updated with the variables.
    """
    updated_context = context
    for name, variable_evaluator in variable_evaluators.items():
        variable_value = variable_evaluator.evaluate(updated_context)
        updated_context = updated_context.with_variables({name: variable_value})

    return updated_context


def _to_result_node(xpath_node: XPathNode) -> XMLNode:
    """Transform the provided XPathNode from the `elementpath` library into on of our XMLNode instances.

    Args:
        xpath_node: the node we would like to wrap in our XMLNodes. This can be text node, processing instruction etc.

    Returns:
        A corresponding instance of our XMLNodes.

    Raises:
        ValueError if we could not match the provided XPathNode.
    """
    path_query = XPath31Parser().parse('path()')
    xpath_location = path_query.evaluate(XPathContext(xpath_node.root_node, item=xpath_node))

    match xpath_node.kind:
        case 'element':
            return ElementNode(xpath_location, xpath_node.elem)
        case 'attribute':
            return AttributeNode(xpath_location, xpath_node.name, xpath_node.value, xpath_node.parent.elem)
        case 'processing-instruction':
            return ProcessingInstructionNode(xpath_location, xpath_node.elem)
        case 'comment':
            return CommentNode(xpath_location, xpath_node.elem)
    raise ValueError(f'Could not transform XPathNode of kind "{xpath_node.kind}".')


def get_subject_node(subject_xpath_expression: XPathExpression | None,
                     query_parser: QueryParser,
                     evaluation_context: EvaluationContext) -> XMLNode | None:
    """Get the node referenced by the optional subject in the rule or check.

    Rules and checks may have a subject indicating an alternative location for the validated content.
    If the node has this, we will return it.

    Args:
        subject_xpath_expression: the expression of the subject attribute.
        query_parser: the parser we can use to evaluate the xpath expression
        evaluation_context: we need the current evaluation context to evaluate the query expression in the referenced
            subject.

    Returns:
        If we have a subject xpath expression, return the corresponding subject node, if found. Else return None.
    """
    if not subject_xpath_expression:
        return None

    subject_xml_node = query_parser.parse(subject_xpath_expression.expression).evaluate(evaluation_context)
    if subject_xml_node:
        if isinstance(subject_xml_node, list):
            return _to_result_node(subject_xml_node[0])
        else:
            return _to_result_node(subject_xml_node)




from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2023-03-24'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'GPL v3'

from abc import ABCMeta, abstractmethod
from typing import Any, Self, override, Callable
from elementpath.tree_builders import RootArgType
from elementpath.xpath_context import ItemArgType


class QueryProcessor(metaclass=ABCMeta):
    """Interface class for the query processing classes.

    Successful query parsing requires a query parser and an evaluation context. These need to be matched to each other.
    This class ensures matching parsers and evaluation contexts.
    """

    @abstractmethod
    def get_query_parser(self) -> QueryParser:
        """Get the query parser for parsing the queries in an AST.

        Returns:
            A query parser to parse queries in the Schematron
        """

    @abstractmethod
    def get_evaluation_context(self) -> EvaluationContext:
        """Get the evaluation context for the parsed queries.

        Returns:
            An evaluation context to evaluate the parsed queries.
        """

    @abstractmethod
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        """Create a copy of this query processor with updated namespaces.

        Args:
            namespaces: a dictionary mapping namespace prefixes to URIs.

        Returns:
            An updated Query Processor.
        """

    @abstractmethod
    def with_custom_function(self, custom_function: CustomQueryFunction) -> Self:
        """Create a copy of this query processor with support for the provided custom function.

        Args:
            custom_function: the custom function to add to the parser.

        Returns:
            An updated query processor
        """


class SimpleQueryProcessor(QueryProcessor):

    def __init__(self, query_parser: QueryParser, evaluation_context: EvaluationContext):
        """Simple query processor prepared with a query parser and evaluation context.

        Defined to be immutable.

        Args:
            query_parser: the query parser this instance specialize in
            evaluation_context: the evaluation context this instance specializes in
        """
        self._query_parser = query_parser
        self._evaluation_context = evaluation_context

    @override
    def get_query_parser(self) -> QueryParser:
        return self._query_parser

    @override
    def get_evaluation_context(self) -> EvaluationContext:
        return self._evaluation_context

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return type(self)(self._query_parser.with_namespaces(namespaces),
                          self._evaluation_context.with_namespaces(namespaces))

    @override
    def with_custom_function(self, custom_function: CustomQueryFunction) -> Self:
        return type(self)(self._query_parser.with_custom_function(custom_function),
                          self._evaluation_context)


class QueryParser(metaclass=ABCMeta):
    """Representation of a parser for Schematron queries."""

    @abstractmethod
    def parse(self, source: str) -> Query:
        """Parse an expression in the implemented query language.

        Args:
            source: the source code of the expression, in the language supported by this parser.

        Returns:
            A parsed expression in the language supported by this parser.
        """

    @abstractmethod
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        """Create a copy of this query parser with updated namespaces.

        Args:
            namespaces: a dictionary mapping namespace prefixes to URIs.

        Returns:
            An updated Query Parser.
        """

    @abstractmethod
    def with_custom_function(self, custom_function: CustomQueryFunction) -> Self:
        """Create a copy of this query parser with an additional custom query function.

        Args:
            custom_function: the custom function to add to the parser.

        Returns:
            An updated Query Parser.
        """


class CachingQueryParser(QueryParser):

    def __init__(self, query_parser: QueryParser):
        """A wrapper around a query parser enabling caching of compiled queries.

        This keeps a mapping of source strings to Queries and checks this first before compiling a query.

        Args:
            query_parser: the query parser we use for actual parsing
        """
        self._query_parser = query_parser
        self._query_cache = {}

    @override
    def parse(self, source: str) -> Query:
        return self._query_cache.setdefault(source, self._query_parser.parse(source))

    @override
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        return type(self)(self._query_parser.with_namespaces(namespaces))

    @override
    def with_custom_function(self, custom_function: CustomQueryFunction) -> Self:
        return type(self)(self._query_parser.with_custom_function(custom_function))


class CustomQueryFunction(metaclass=ABCMeta):

    @property
    @abstractmethod
    def callback(self) -> Callable[..., Any]:
        """Get the callback of this custom function.

        Returns:
            The (Python) callback function.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of this function.

        Returns:
            The name of the callable function.
        """

    @property
    @abstractmethod
    def prefix(self) -> str | None:
        """The XML prefix of this function.

        Returns:
            The XML prefix of this function
        """


class SimpleCustomQueryFunction(CustomQueryFunction):

    def __init__(self, callback: Callable[..., Any], name: str, prefix: str | None = None):
        """Simple definition of a custom query function.

        Args:
            callback: the function to call when evaluating the parsed expression
            name: the function name for use inside the query language.
            prefix: the function's name prefix, if not provided it is set to the XPath default
                function namespace (`fn:`). This means, it may overwrite library functions.
        """
        self._callback = callback
        self._name = name
        self._prefix = prefix

    @property
    def callback(self) -> Callable[..., Any]:
        return self._callback

    @property
    def name(self) -> str:
        return self._name

    @property
    def prefix(self) -> str | None:
        return self._prefix


class EvaluationContext(metaclass=ABCMeta):
    """Representation of the context required when evaluating a Query.

    Each context should be immutable. Every change constructs a new evaluation context.
    """

    @abstractmethod
    def with_xml_root(self, xml_root: RootArgType) -> Self:
        """Create a new evaluation context with the XML root node we can use for dynamic queries.

        For queries like: `xs:integer(42)` no XML root node is needed.
        For dynamic queries like: `xs:integer(/data/@nmr_items)`, a root node is needed.

        Args:
            xml_root: the root node usable for dynamic query evaluations

        Returns:
            A new evaluation context
        """

    @abstractmethod
    def with_context_item(self, xml_item: ItemArgType) -> Self:
        """Create a new evaluation context with the provided xml item (node, comment, attribute) as query base.

        This is needed for asserts and reports queries which assume the context of the rule node.

        Args:
            xml_item: the XML item we use as context node for parser evaluation.

        Returns:
            A new evaluation context
        """

    @abstractmethod
    def with_namespaces(self, namespaces: dict[str, str]) -> Self:
        """Create a new evaluation context with the namespaces used during evaluation.

        Args:
            namespaces: a dictionary mapping namespace prefixes to URIs.
                This is used when namespace information is not available within document and element nodes.

        Returns:
            A new evaluation context
        """

    @abstractmethod
    def with_variables(self, variables: dict[str, Any], overwrite: bool = False) -> Self:
        """Create a new evaluation context with the namespaces used during evaluation.

        Args:
            variables: a dictionary mapping variable names (QNames) to variables. This expects the
                variables to be a parsed and evaluated value.
            overwrite: if set to True, we will overwrite any stored variables. If set to False, we update
                the dictionary of variables.

        Returns:
            A new evaluation context
        """

    @abstractmethod
    def get_xml_root(self) -> RootArgType | None:
        """Get the XML root node current in this context.

        Returns:
            The current root node. May not be set, giving None.
        """

    @abstractmethod
    def get_context_item(self) -> ItemArgType | None:
        """Get the XML node serving as the query base.

        Returns:
            The XML item we use as context node for parser evaluation. May not be set yet.
        """


class Query(metaclass=ABCMeta):
    """Representation of an executable Schematron query.

    To specialize for a new language, one must implement a specialized Context, Parser and Query.
    """

    @abstractmethod
    def evaluate(self, context: EvaluationContext | None = None) -> Any:
        """Evaluate this query.

        Args:
            context: optional context to be used during evaluation.
                The exact context and its usage is implementation defined.

        Returns:
            The results of running this query.
        """



__author__ = 'Robbert Harms'
__date__ = '2023-04-24'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'GPL v3'

from abc import ABCMeta, abstractmethod

from lxml.etree import Element, SubElement, ElementTree, _ElementTree

from datetime import datetime

from pyschematron import __version__
from pyschematron.direct_mode.schematron.ast import Assert, Namespace, ConcretePattern
from pyschematron.direct_mode.svrl.ast import ActivePattern, FiredRule, SchematronQuery, FailedAssert, \
    XPathExpression, Text, SchematronOutput, SuccessfulReport, SuppressedRule, NSPrefixInAttributeValues, MetaData, \
    CheckResult, ValidationEvent, PropertyReference, DiagnosticReference

from pyschematron.direct_mode.svrl.xml_writer import LxmlSVRLWriter
from pyschematron.direct_mode.xml_validation.results.validation_results import (XMLDocumentValidationResult, RuleResult,
                                                                                FiredRuleResult, SuppressedRuleResult,
                                                                                PropertyResult, DiagnosticResult)


class SVRLReportBuilder(metaclass=ABCMeta):

    @abstractmethod
    def create_svrl_xml(self, validation_result: XMLDocumentValidationResult) -> _ElementTree:
        """Create a Schematron Validation Reporting Language (SVRL) document of the validation results.

        This transforms the validation results into an XML document in the SVRL namespace.

        Args:
            validation_result: the result of validating the document using the validator of this package.

        Returns:
            An XML document in the SVRL namespace.
        """


class DefaultSVRLReportBuilder(SVRLReportBuilder):
    """Converts the validation results into an SVRL AST.

    In the SVRL report we want to enumerate the processing results in the format:
        <active-pattern/>
        ...
        <rule results>
        ...
        <active-pattern/>
        ...

    The results of using the validator provides results for every XML node, Schematron pattern and Schematron rule
    checked. This is too comprehensive for the SVRL and as such this class simplifies the results.
    """

    def create_svrl_xml(self, validation_result: XMLDocumentValidationResult) -> _ElementTree:
        title = None
        if title_node := validation_result.schema_information.schema.title:
            title = title_node.content

        svrl = SchematronOutput(
            self._get_text_nodes(validation_result),
            self._get_ns_prefix_nodes(validation_result),
            tuple(self.get_validation_events(validation_result)),
            metadata=self._get_metadata(validation_result),
            phase=validation_result.schema_information.phase,
            schema_version=validation_result.schema_information.schema.schema_version,
            title=title)

        writer = LxmlSVRLWriter()
        return ElementTree(writer.create_xml(svrl))

    def get_validation_events(self, validation_result: XMLDocumentValidationResult) -> list[ValidationEvent]:
        """Extract a list of SVRL validation events from the validation results.

        Args:
            validation_result: The validation results from the validator

        Returns:
            A list of SVRL validation events (active pattern, fired rule, failed assert,
            successful report, suppressed rule).
        """
        root_document = None
        if xml_document := validation_result.xml_information.xml_document:
            if docinfo := xml_document.docinfo:
                root_document = docinfo.URL

        processed_patterns = self._get_rule_results_by_pattern(validation_result)
        validation_events = []
        for pattern, rules_processed in processed_patterns.items():
            name = pattern.title.content if pattern.title else None
            validation_events.append(ActivePattern(id=pattern.id, name=name, documents=(root_document,)))

            for rule_processed in rules_processed:
                if isinstance(rule_processed, FiredRuleResult):
                    validation_events.append(FiredRule(SchematronQuery(rule_processed.rule.context.query),
                                                       id=rule_processed.rule.id))
                    validation_events += self._get_svrl_check_results(rule_processed)
                elif isinstance(rule_processed, SuppressedRuleResult):
                    validation_events.append(SuppressedRule(SchematronQuery(rule_processed.rule.context.query)))
        return validation_events

    def _get_svrl_check_results(self, fired_rule: FiredRuleResult) -> list[CheckResult]:
        """Transform the validation checks in the fired rule to a list of SVRL AST check results.

        Args:
            fired_rule: the rule which was fired, from this we will return the checks which did not succeed.

        Returns:
            A list of failed asserts or successful reports.
        """
        subject_location = None
        if fired_rule.subject_node:
            subject_location = XPathExpression(fired_rule.subject_node.xpath_location)

        events = []
        for check_result in fired_rule.check_results:
            if check_result.check_result:
                if check_result.subject_node:
                    subject_location = XPathExpression(check_result.subject_node.xpath_location)

                property_references = self._get_property_references(check_result.property_results)
                diagnostic_references = self._get_diagnostic_reference(check_result.diagnostic_results)

                if isinstance(check_result.check, Assert):
                    event = FailedAssert(Text(check_result.text),
                                         XPathExpression(check_result.xml_node.xpath_location),
                                         SchematronQuery(check_result.check.test.query),
                                         diagnostic_references=tuple(diagnostic_references),
                                         property_references=tuple(property_references),
                                         subject_location=subject_location)
                else:
                    event = SuccessfulReport(Text(check_result.text),
                                             XPathExpression(check_result.xml_node.xpath_location),
                                             SchematronQuery(check_result.check.test.query),
                                             diagnostic_references=tuple(diagnostic_references),
                                             property_references=tuple(property_references),
                                             subject_location=subject_location)
                events.append(event)
        return events

    @staticmethod
    def _get_property_references(property_results: list[PropertyResult] | None) -> list[PropertyReference]:
        """Convert the property results to SVRL property reference nodes.

        Args:
            property_results: the list of property results, may be None, in which case we return an empty list.

        Returns:
            The list of property references.
        """
        property_references = []
        if property_results:
            for property_result in property_results:
                property_references.append(PropertyReference(
                    Text(property_result.text), property_result.property_id,
                    property_result.role, property_result.scheme))
        return property_references

    @staticmethod
    def _get_diagnostic_reference(diagnostic_results: list[DiagnosticResult] | None) -> list[DiagnosticReference]:
        """Convert the diagnostic results to SVRL diagnostic reference nodes.

        Args:
            diagnostic_results: the list of diagnostic results, may be None, in which case we return an empty list.

        Returns:
            The list of diagnostic references.
        """
        diagnostic_references = []
        if diagnostic_results:
            for diagnostic_result in diagnostic_results:
                text = Text(diagnostic_result.text,
                            xml_lang=diagnostic_result.xml_lang,
                            xml_space=diagnostic_result.xml_space)
                diagnostic_references.append(DiagnosticReference(text, diagnostic=diagnostic_result.diagnostic_id))
        return diagnostic_references

    @staticmethod
    def _get_rule_results_by_pattern(
            validation_result: XMLDocumentValidationResult) -> dict[ConcretePattern, list[RuleResult]]:
        """Reduce the validation results into a dictionary indexed by pattern.

        The validation results consist of data for each XML node, Schematron pattern and Schematron rule. This function
        iterates over the validation results and groups them by Schematron pattern.

        Args:
            validation_result: the validation results

        Returns:
            A dictionary which as indices the patterns which had at least one fired rule. As values are the
            rule results taken from the validation results.
        """
        processed_patterns = {}
        for node_result in validation_result.node_results:
            for pattern_result in node_result.pattern_results:
                rules_processed = processed_patterns.setdefault(pattern_result.pattern, [])

                if pattern_result.has_fired_rule():
                    for rule_result in pattern_result.rule_results:
                        rules_processed.append(rule_result)
        return processed_patterns

    @staticmethod
    def _get_text_nodes(validation_result: XMLDocumentValidationResult) -> tuple[Text, ...]:
        """Get the listing of text nodes we will add to the SVRL output.

        For the text nodes we will use all the paragraph nodes present in the Schematron document.

        Args:
            validation_result: all the validation information

        Returns:
            The text nodes, created from the paragraphs of the Schematron document.
        """
        texts = []
        for paragraph in validation_result.schema_information.schema.paragraphs:
            texts.append(Text(paragraph.content, icon=paragraph.icon,
                              xml_lang=paragraph.xml_lang, id=paragraph.id,
                              class_=paragraph.class_))
        return tuple(texts)

    @staticmethod
    def _get_ns_prefix_nodes(validation_result: XMLDocumentValidationResult) -> tuple[NSPrefixInAttributeValues, ...]:
        """Get the listing of the ns prefix nodes.

        Args:
            validation_result: all the validation information

        Returns:
            The NS prefix nodes.
        """
        ns_prefix_in_attribute_values = []
        for schematron_namespace in validation_result.schema_information.schema.namespaces:
            ns_prefix_in_attribute_values.append(NSPrefixInAttributeValues(schematron_namespace.prefix,
                                                                           schematron_namespace.uri))
        return tuple(ns_prefix_in_attribute_values)

    @staticmethod
    def _get_metadata(validation_result: XMLDocumentValidationResult) -> MetaData:
        """Get the metadata node to add to the SVRL.

        Args:
            validation_result: all the validation information

        Returns:
            The metadata node.
        """
        namespaces = (Namespace('dct', 'http://purl.org/dc/terms/'),
                      Namespace('skos', 'http://www.w3.org/2004/02/skos/core#'),
                      Namespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'),
                      Namespace('pysch', 'https://github.com/robbert-harms/pyschematron'))

        nsmap = {ns.prefix: ns.uri for ns in namespaces}
        create_time = datetime.now().astimezone()

        def creator_element():
            creator = Element(f'{{{nsmap["dct"]}}}creator', nsmap=nsmap)
            agent = SubElement(creator, f'{{{nsmap["dct"]}}}agent')
            pref_label = SubElement(agent, f'{{{nsmap["skos"]}}}prefLabel')
            pref_label.text = f'PySchematron {__version__}'
            return creator

        def created_element():
            created = Element(f'{{{nsmap["dct"]}}}created', nsmap=nsmap)
            created.text = create_time.isoformat()
            return created

        def source_element():
            source = Element(f'{{{nsmap["dct"]}}}source', nsmap=nsmap)
            description = SubElement(source, f'{{{nsmap["rdf"]}}}Description')
            creator = SubElement(description, f'{{{nsmap["dct"]}}}creator')
            agent = SubElement(creator, f'{{{nsmap["dct"]}}}Agent')
            pref_label = SubElement(agent, f'{{{nsmap["skos"]}}}prefLabel')
            pref_label.text = f'PySchematron {__version__}'
            description.append(created_element())
            return source

        return MetaData((creator_element(), created_element(), source_element()), namespaces)



__author__ = 'Robbert Harms'
__date__ = '2023-03-06'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

import re


def macro_expand(string: str, macros: dict[str, str]) -> str:
    """Expand the provided macros on the provided string.

    This replaces all the macros in one go. This is a specialized version of multi-string replacement which
    assumes all the replacements are unique (as they are in macro's).

    We use this function for instantiating abstract patterns to concrete patterns.

    This assumes the macros already have the prefix `$`.

    Args:
        string: the string on which to apply the macro
        macros: the macros to expand

    Returns:
        A version of the string with the macros expanded
    """
    macros_pattern = '|'.join(re.escape(k) for k in macros)
    pattern = re.compile(f'({macros_pattern})\\b')
    return pattern.sub(lambda match: macros[match.group(0)], string)



from __future__ import annotations

__author__ = 'Robbert Harms'
__date__ = '2023-03-25'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'GPL v3'

from abc import ABCMeta, abstractmethod
from typing import override

from pyschematron.direct_mode.schematron.ast import Schema
from pyschematron.direct_mode.xml_validation.queries.base import QueryProcessor
from pyschematron.direct_mode.xml_validation.queries.xpath import XPath1QueryParser, XPath2QueryParser, \
    XPath3QueryParser, XPath31QueryParser, XPathQueryProcessor


class QueryProcessorFactory(metaclass=ABCMeta):
    """Query processor factories can construct QueryProcessor classes specific to your query binding language.

    In Schematron, the queryBinding attribute determines which query language is used. This factory
    allows you to get the right query processor for your query binding language.
    """

    @abstractmethod
    def get_query_processor(self, query_binding: str) -> QueryProcessor:
        """Get the processor you can use for this query binding language.

        Args:
            query_binding: the query binding for which we want to get a parser.

        Returns:
            A query processor specialized for this query binding language.

        Raises:
            ValueError: if no query processor could be found for the indicated query binding.
        """

    @abstractmethod
    def has_query_processor(self, query_binding: str) -> bool:
        """Check if we have a processor for the specific query binding language.

        Args:
            query_binding: the query binding for which we want to check if a processor is available.

        Returns:
            True if we have a processor, False otherwise.
        """

    @abstractmethod
    def get_schema_query_processor(self, schema: Schema) -> QueryProcessor:
        """Get the processor you can use for this schema.

        Not only will this select the right query binding, it will also load the namespaces.

        Args:
            schema: the Schema for which we want to get a query processor.

        Returns:
            A query processor specialized for this Schema, with the right query binding language and
                the namespaces loaded.

        Raises:
            ValueError: if no query processor could be found for this Schema.
        """


class DefaultQueryProcessorFactory(QueryProcessorFactory):

    def __init__(self):
        """The default query processor factory.

        This factory only supports XSLT and XPath query languages. The XSLT query binding is additionally limited
        to XPath expressions.
        """
        self._query_processors = {
            'xslt': XPathQueryProcessor(XPath1QueryParser()),
            'xslt2': XPathQueryProcessor(XPath2QueryParser()),
            'xslt3': XPathQueryProcessor(XPath3QueryParser()),
            'xpath': XPathQueryProcessor(XPath1QueryParser()),
            'xpath2': XPathQueryProcessor(XPath2QueryParser()),
            'xpath3': XPathQueryProcessor(XPath3QueryParser()),
            'xpath31': XPathQueryProcessor(XPath31QueryParser()),
        }

    @override
    def get_query_processor(self, query_binding: str) -> QueryProcessor:
        try:
            return self._query_processors[query_binding]
        except KeyError:
            raise ValueError(f'No parser could be found for the query binding "{query_binding}".')

    @override
    def has_query_processor(self, query_binding: str) -> bool:
        return query_binding in self._query_processors

    @override
    def get_schema_query_processor(self, schema: Schema) -> QueryProcessor:
        query_binding = schema.query_binding or 'xslt'
        namespaces = {ns.prefix: ns.uri for ns in schema.namespaces}

        processor = self.get_query_processor(query_binding)
        return processor.with_namespaces(namespaces)


class ExtendableQueryProcessorFactory(DefaultQueryProcessorFactory):

    def __init__(self):
        """An extendable query processor factory.

        This has all the processors from the default query processor factory, but allows extending and/or overwriting
        these using getters and setters.
        """
        super().__init__()

    def set_query_processor(self, query_binding: str, query_processor: QueryProcessor):
        """Set the query processor to use for a specific query binding language.

        Args:
            query_binding: the query binding we wish to add / overwrite.
            query_processor: the query processor we would like to use for this query binding.
        """
        self._query_processors[query_binding] = query_processor



__author__ = 'Robbert Harms'
__date__ = '2022-02-25'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

from pathlib import Path
from typing import Literal

from lxml.etree import _ElementTree

from pyschematron.__version__ import __version__
from pyschematron.api import ValidationResult
from pyschematron.direct_mode.api import DirectModeSchematronValidatorFactory
from pyschematron.direct_mode.xml_validation.queries.base import CustomQueryFunction


def validate_document(xml_document: Path | _ElementTree,
                      schematron_schema: Path | _ElementTree,
                      phase: str | None = None,
                      custom_functions: dict[str, str | list[CustomQueryFunction]] | None = None,
                      mode: Literal['direct-mode'] = 'direct-mode') -> ValidationResult:
    """Validate an XML document using a Schematron schema.

    Args:
        xml_document: the XML document we would like to validate
        schematron_schema: the Schematron Schema we would like to load and use for the validation
        phase: the Schematron phase we would like to use, optional.
        custom_functions: a dictionary defining additional custom functions to add to the parser(s).
            This should at least contain the key 'query_binding' mapping to a query binding name, and the
            key 'custom_query_functions' specifying a list of custom query functions to add.
            For non-standard query binding language, you also need to provide the key 'base_query_binding'
            mapping to a standard query binding. Example usage:
            `{'query_binding': 'xpath31-custom', 'base_query_binding': 'xpath31', 'custom_query_functions': [...]}`.
        mode: which validation mode we would like to use, at the moment this only supports the 'direct-mode'.

    Returns:
        The validation result in an API wrapper.
    """
    validator_factory = DirectModeSchematronValidatorFactory(schematron_xml=schematron_schema, phase=phase)
    if custom_functions:
        validator_factory.add_custom_functions(**custom_functions)
    validator = validator_factory.build()
    return validator.validate(xml_document)


def validate_documents(xml_documents: list[Path | _ElementTree],
                       schematron_schema: Path | _ElementTree,
                       phase: str | None = None,
                       custom_functions: dict[str, str | list[CustomQueryFunction]] | None = None,
                       mode: Literal['direct-mode'] = 'direct-mode') -> list[ValidationResult]:
    """Validate multiple XML documents using the same Schematron schema.

    This assumes we would like to use the same Schematron phase for each validation. As such, we can afford a speed-up
    since we don't need to compile the Schematron every run again.

    Args:
        xml_documents: the XML documents we would like to validate
        schematron_schema: the Schematron Schema we would like to load and use for the validation
        phase: the Schematron phase we would like to use, optional.
        custom_functions: a dictionary defining additional custom functions to add to the parser(s).
            This should at least contain the key 'query_binding' mapping to a query binding name, and the
            key 'custom_query_functions' specifying a list of custom query functions to add.
            For non-standard query binding language, you also need to provide the key 'base_query_binding'
            mapping to a standard query binding. Example usage:
            `{'query_binding': 'xpath31-custom', 'base_query_binding': 'xpath31', 'custom_query_functions': [...]}`.
        mode: which validation mode we would like to use, at the moment this only supports the 'direct-mode'.

    Returns:
        The validation results in an API wrapper.
    """
    validator_factory = DirectModeSchematronValidatorFactory(schematron_xml=schematron_schema, phase=phase)
    if custom_functions:
        validator_factory.add_custom_functions(**custom_functions)
    validator = validator_factory.build()

    validation_results = []
    for document in xml_documents:
        validation_results.append(validator.validate(document))

    return validation_results





"""A test script I use when developing PySchematron."""

__author__ = 'Robbert Harms'
__date__ = '2023-02-21'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'

from pathlib import Path

from lxml import etree

from pyschematron.direct_mode.schematron.ast_visitors import ResolveExtendsVisitor, \
    ResolveAbstractPatternsVisitor, PhaseSelectionVisitor
from pyschematron.direct_mode.schematron.parsers.xml.parser import ParsingContext, SchemaParser
from pyschematron.direct_mode.xml_validation.results.svrl_builder import DefaultSVRLReportBuilder
from pyschematron.direct_mode.xml_validation.validators import SimpleSchematronXMLValidator
from pyschematron.utils import load_xml_document

'''
cd programming/python/pyschematron/tests/fixtures/full_example/
java -jar ~/programming/java/schxslt-cli.jar -d cargo.xml -s schema.sch -o /tmp/report.xml
'''

example_path = Path('../tests/fixtures/full_example/')
schematron_path = example_path / 'schema.sch'
phase = '#ALL'

schematron_xml = load_xml_document(schematron_path)
parsing_context = ParsingContext(base_path=schematron_path.parent)

schematron_parser = SchemaParser()
schema = schematron_parser.parse(schematron_xml.getroot(), parsing_context)
schema = ResolveExtendsVisitor(schema).apply(schema)
schema = ResolveAbstractPatternsVisitor(schema).apply(schema)
schema = PhaseSelectionVisitor(schema, phase).apply(schema)

validator = SimpleSchematronXMLValidator(schema, phase, parsing_context.base_path)

xml_document = load_xml_document(example_path / 'cargo.xml')
validation_results = validator.validate_xml(xml_document)

svrl_report = DefaultSVRLReportBuilder().create_svrl_xml(validation_results)
report_str = etree.tostring(svrl_report, pretty_print=True).decode('utf-8')

with open('/tmp/report_pyschematron.xml', 'w') as f:
    f.write(report_str)

print(report_str)

print()



"""This script demonstrates using the direct-mode PySchematron validator to validate your XML documents.

In Schematron validation, we apply a Schematron Schema to an XML resulting in either a pass or a fail. A fail indicates
that the document could not be validated using the Schema, hence the XML may have problems. In addition to this boolean
output, Schematron also defines the Schematron Validation Report Language (SVRL), loosely defining a format in which
more information about the validation results can be represented.

This script shows three different ways of interacting with the PySchematron direct-mode validator. The most simple is
by using the functional interface defined in the main module. Second, you can use a generalized API which might be
extended in the future with an XSLT methodology. Finally, you can use the full-blown direct-mode classes and methods.
The latter is the most complicated but gives the most control.
"""

__author__ = 'Robbert Harms'
__date__ = '2024-04-03'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from pathlib import Path

from lxml import etree

from pyschematron import DirectModeSchematronValidatorFactory, validate_document
from pyschematron.direct_mode.schematron.ast_visitors import ResolveExtendsVisitor, ResolveAbstractPatternsVisitor, \
    PhaseSelectionVisitor
from pyschematron.direct_mode.schematron.parsers.xml.parser import SchemaParser, ParsingContext
from pyschematron.direct_mode.xml_validation.results.svrl_builder import DefaultSVRLReportBuilder
from pyschematron.direct_mode.xml_validation.validators import SimpleSchematronXMLValidator
from pyschematron.utils import load_xml_document


# the paths to the example data and Schema
example_base_path = Path('../tests/fixtures/full_example/')
schematron_schema_path = example_base_path / 'schema.sch'
example_xml_document_path = example_base_path / 'cargo.xml'

# the phase we would like to evaluate
phase = '#ALL'


def demo_functional_interface():
    """This example uses the functional interface, the most simple method of interacting with PySchematron. """
    result = validate_document(example_xml_document_path, schematron_schema_path)
    svrl = result.get_svrl()

    report_str = etree.tostring(svrl, pretty_print=True).decode('utf-8')
    print(report_str)
    print(result.is_valid())


def demo_generic_api():
    """This demonstrates the use of the generic API."""
    validator_factory = DirectModeSchematronValidatorFactory()
    validator_factory.set_schema(schematron_schema_path)
    validator_factory.set_phase(phase)

    validator = validator_factory.build()
    validation_result = validator.validate(example_xml_document_path)

    svrl = validation_result.get_svrl()
    report_str = etree.tostring(svrl, pretty_print=True).decode('utf-8')

    print(report_str)
    print(validation_result.is_valid())


def demo_full_api():
    """This demonstrates the inner workings of the direct-mode validator."""
    schematron_xml = load_xml_document(schematron_schema_path)
    parsing_context = ParsingContext(base_path=schematron_schema_path.parent)

    schematron_parser = SchemaParser()
    schema = schematron_parser.parse(schematron_xml.getroot(), parsing_context)
    schema = ResolveExtendsVisitor(schema).apply(schema)
    schema = ResolveAbstractPatternsVisitor(schema).apply(schema)
    schema = PhaseSelectionVisitor(schema, phase).apply(schema)

    validator = SimpleSchematronXMLValidator(schema, phase, parsing_context.base_path)

    xml_document = load_xml_document(example_xml_document_path)
    validation_results = validator.validate_xml(xml_document)

    svrl_report = DefaultSVRLReportBuilder().create_svrl_xml(validation_results)

    report_str = etree.tostring(svrl_report, pretty_print=True).decode('utf-8')
    print(report_str)
    print(validation_results.is_valid())


demo_functional_interface()
demo_generic_api()
demo_full_api()

